# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from collections import defaultdict
from collections import namedtuple
from collections import OrderedDict
import copy
import re
import time

from oslo_config import cfg
from oslo_log import log

from vitrage.common.constants import EdgeProperties as EProps
from vitrage.common.constants import VertexProperties as VProps
from vitrage.common.utils import md5
from vitrage.common.utils import recursive_keypairs
from vitrage.entity_graph.mappings.datasource_info_mapper \
    import DatasourceInfoMapper
from vitrage.evaluator.actions.action_executor import ActionExecutor
from vitrage.evaluator.actions.base import ActionMode
from vitrage.evaluator.actions.base import ActionType
import vitrage.evaluator.actions.priority_tools as pt
from vitrage.evaluator.template_data import ActionSpecs
from vitrage.evaluator.template_data import EdgeDescription
from vitrage.evaluator.template_functions.function_resolver import is_function
from vitrage.evaluator.template_schema_factory import TemplateSchemaFactory
from vitrage.graph.algo_driver.algorithm import Mapping
from vitrage.graph.algo_driver.sub_graph_matching import \
    NEG_CONDITION
from vitrage.graph.driver import Vertex
from vitrage import storage
from vitrage.storage.sqlalchemy import models
from vitrage.utils.datetime import utcnow

CONF = cfg.CONF
LOG = log.getLogger(__name__)

# Entry containing action info.
# specs - ActionSpecs
# mode - DO or UNDO (the action)
# action_id - the action id in scenario_repository
# Trigger_id  - a unique identifier per match in graph (i.e., the subgraph
# that matched the action in the spec) for the specific action.
ActionInfo = \
    namedtuple('ActionInfo', ['specs', 'mode', 'action_id', 'trigger_id'])

TARGET = 'target'
SOURCE = 'source'


class ScenarioEvaluator(object):

    def __init__(self,
                 e_graph,
                 scenario_repo,
                 actions_callback,
                 enabled=False):
        self._entity_graph = e_graph
        self._db = storage.get_connection_from_config()
        self._scenario_repo = scenario_repo
        self._action_executor = ActionExecutor(actions_callback)
        self._entity_graph.subscribe(self.process_event)
        self.enabled = enabled
        self.connected_component_cache = defaultdict(dict)

    @property
    def scenario_repo(self):
        return self._scenario_repo

    @scenario_repo.setter
    def scenario_repo(self, scenario_repo):
        self._scenario_repo = scenario_repo

    def run_evaluator(self, action_mode=ActionMode.DO):
        self.enabled = True
        vertices = self._entity_graph.get_vertices()
        start_time = time.time()
        for vertex in vertices:
            if action_mode == ActionMode.DO:
                self.process_event(None, vertex, True)
            elif action_mode == ActionMode.UNDO:
                self.process_event(vertex, None, True)
        LOG.info(
            'Run %s Evaluator on %s items - took %s',
            action_mode, len(vertices), (time.time() - start_time))

    def process_event(self, before, current, is_vertex, *args, **kwargs):
        """Notification of a change in the entity graph.

        :param is_vertex:
        :param before: The graph element (vertex or edge) prior to the
        change that happened. None if the element was just created.
        :param current: The graph element (vertex or edge) after the
        change that happened. Deleted elements should arrive with the
        vitrage_is_deleted property set to True
        """

        if not self.enabled:
            LOG.debug("Process event disabled")
            return

        LOG.debug('Process event - starting')
        LOG.debug("Element before event: %s, Current element: %s",
                  before,
                  current)

        before_scenarios = self._get_element_scenarios(before, is_vertex)
        current_scenarios = self._get_element_scenarios(current, is_vertex)
        before_scenarios, current_scenarios = \
            self._remove_overlap_scenarios(before_scenarios, current_scenarios)

        if len(before_scenarios) + len(current_scenarios):
            LOG.debug("Number of relevant scenarios found: undo = %s, do = %s",
                      len(before_scenarios),
                      len(current_scenarios))

        actions = self._process_and_get_actions(before,
                                                before_scenarios,
                                                ActionMode.UNDO)
        actions.extend(self._process_and_get_actions(current,
                                                     current_scenarios,
                                                     ActionMode.DO))
        actions_to_preform = []
        try:
            actions_to_preform = self._analyze_and_filter_actions(actions)
        except Exception:
            LOG.exception("Evaluator error, will not execute actions %s",
                          actions)

        self._action_executor.execute(actions_to_preform)
        LOG.debug('Process event - completed')

    def _get_element_scenarios(self, element, is_vertex):
        if not element \
                or element.get(VProps.VITRAGE_IS_DELETED) \
                or element.get(EProps.VITRAGE_IS_DELETED):
            return []
        elif is_vertex:
            return self._scenario_repo.get_scenarios_by_vertex(element)
        else:  # is edge
            edge_desc = self._get_edge_description(element)
            return self._scenario_repo.get_scenarios_by_edge(edge_desc)

    def _get_edge_description(self, element):
        source = self._entity_graph.get_vertex(element.source_id)
        target = self._entity_graph.get_vertex(element.target_id)
        edge_desc = EdgeDescription(element, source, target)
        return edge_desc

    @staticmethod
    def _remove_overlap_scenarios(before, current):
        intersection = list(filter(lambda x: x in before, current))
        before = list(filter(lambda x: x not in intersection, before))
        current = list(filter(lambda x: x not in intersection, current))
        return before, current

    def _process_and_get_actions(self, element, triggered_scenarios, mode):
        actions = []
        for triggered_scenario in triggered_scenarios:
            LOG.debug("Processing: %s", triggered_scenario)
            scenario_element = triggered_scenario[0]
            scenario = triggered_scenario[1]
            actions.extend(self._process_scenario(element,
                                                  scenario,
                                                  scenario_element,
                                                  mode))
        return actions

    def _process_scenario(self, element, scenario, scenario_elements, mode):
        if not isinstance(scenario_elements, list):
            scenario_elements = [scenario_elements]
        actions = []
        for action in scenario.actions:
            for scenario_element in scenario_elements:
                matches = self._evaluate_subgraphs(scenario.subgraphs,
                                                   element,
                                                   scenario_element,
                                                   action.targets[TARGET])

                actions.extend(self._get_actions_from_matches(scenario.version,
                                                              matches,
                                                              mode,
                                                              action))

        return actions

    def _evaluate_subgraphs(self,
                            subgraphs,
                            element,
                            scenario_element,
                            action_target):
        if isinstance(element, Vertex):
            return self._find_vertex_subgraph_matching(subgraphs,
                                                       action_target,
                                                       element,
                                                       scenario_element)
        else:
            return self._find_edge_subgraph_matching(subgraphs,
                                                     action_target,
                                                     element,
                                                     scenario_element)

    def _get_actions_from_matches(self,
                                  scenario_version,
                                  combined_matches,
                                  mode,
                                  action_spec):
        actions = []
        for is_switch_mode, matches in combined_matches:
            new_mode = mode
            if is_switch_mode:
                new_mode = ActionMode.UNDO \
                    if mode == ActionMode.DO else ActionMode.DO

            template_schema = \
                TemplateSchemaFactory().template_schema(scenario_version)

            for match in matches:
                match_action_spec = self._get_action_spec(action_spec, match)
                items_ids = \
                    [match_item[1].vertex_id for match_item in match.items()]
                match_hash = md5(tuple(sorted(items_ids)))
                self._evaluate_property_functions(template_schema, match,
                                                  match_action_spec.properties)

                actions.append(ActionInfo(match_action_spec, new_mode,
                                          match_action_spec.id, match_hash))

        return actions

    def _evaluate_property_functions(self, template_schema, match,
                                     action_props):
        """Evaluate the action properties, in case they contain functions

        In template version 2 we introduced functions, and specifically the
        get_attr function. This method evaluate its value and updates the
        action properties, before the action is being executed.

        Example:

        - action:
            action_type: execute_mistral
            properties:
              workflow: evacuate_vm
              input:
                vm_name: get_attr(instance1,name)
                force: false

        In this example, the method will iterate over 'properties', and then
        recursively over 'input', and for 'vm_name' it will replace the
        call for get_attr with the actual name of the VM. The input for the
        Mistral workflow will then be:
        vm_name: vm_1
        force: false

        """
        for key, value in action_props.items():
            if isinstance(value, dict):
                # Recursive call for a dictionary
                self._evaluate_property_functions(template_schema,
                                                  match, value)

            elif value is not None and is_function(value):
                # The value is a function
                func_and_args = re.split('[(),]', value)
                func_name = func_and_args.pop(0)
                args = [arg.strip() for arg in func_and_args if len(arg) > 0]

                # Get the function, execute it and update the property value
                func = template_schema.functions.get(func_name)
                action_props[key] = func(match, *args)

                LOG.debug('Changed property %s value from %s to %s', key,
                          value, action_props[key])

    @staticmethod
    def _get_action_spec(action_spec, match):
        targets = action_spec.targets
        real_items = {
            target: match[target_id] for target, target_id in targets.items()
        }
        return ActionSpecs(action_spec.id,
                           action_spec.type,
                           real_items,
                           copy.deepcopy(action_spec.properties))

    @staticmethod
    def _generate_action_id(action_spec):
        """Generate a unique action id for the action

        BEWARE: The value created here should not be stored in database,
        as in python3, the hash function seed changes after program restart
        """
        targets = [(k, v.vertex_id) for k, v in action_spec.targets.items()]
        return hash(
            (action_spec.type,
             tuple(sorted(targets)),
             tuple(sorted(recursive_keypairs(action_spec.properties))))
        )

    def _analyze_and_filter_actions(self, actions):
        LOG.debug("Actions before filtering: %s", actions)
        if not actions:
            return []

        active_actions = ActiveActionsTracker(self._db, actions)
        for action_info in actions:
            if action_info.mode == ActionMode.DO:
                active_actions.calc_do_action(action_info)
            elif action_info.mode == ActionMode.UNDO:
                active_actions.calc_undo_action(action_info)

        active_actions.flush_db_updates()

        unique_ordered_actions = OrderedDict()
        for action in active_actions.actions_to_perform:
            if isinstance(action, models.ActiveAction):
                action = self._db_action_to_action_info(action)
            id_ = self._generate_action_id(action.specs)
            unique_ordered_actions[id_] = action
        return unique_ordered_actions.values()

    def _find_vertex_subgraph_matching(self,
                                       subgraphs,
                                       action_target,
                                       vertex,
                                       scenario_vertex):
        """calculates subgraph matching for vertex

        iterates over all the subgraphs, and checks if the triggered vertex is
        in the same connected component as the action then run subgraph
        matching on the vertex and return its result, otherwise return an
        empty list of matches.
        """

        matches = []
        for subgraph in subgraphs:
            connected_component = self.get_connected_component(subgraph,
                                                               action_target)

            is_switch_mode = \
                connected_component.get_vertex(scenario_vertex.vertex_id)

            if is_switch_mode:
                initial_map = Mapping(scenario_vertex, vertex, True)
                mat = self._entity_graph.algo.sub_graph_matching(subgraph,
                                                                 initial_map)
                matches.append((False, mat))
            else:
                matches.append((True, []))
        return matches

    def _find_edge_subgraph_matching(self,
                                     subgraphs,
                                     action_target,
                                     edge,
                                     scenario_edge):
        """calculates subgraph matching for edge

        iterates over all the subgraphs, and checks if the triggered edge is a
        negative edge then mark it as deleted=false and negative=false so that
        subgraph matching on that edge will work correctly. after running
        subgraph matching, we need to remove the negative vertices that were
        added due to the change above.
        """

        matches = []
        for subgraph in subgraphs:
            subgraph_edge = subgraph.get_edge(scenario_edge.source.vertex_id,
                                              scenario_edge.target.vertex_id,
                                              scenario_edge.edge.label)
            if not subgraph_edge:
                continue

            is_switch_mode = subgraph_edge.get(NEG_CONDITION, False)

            connected_component = self.get_connected_component(subgraph,
                                                               action_target)
            # change the vitrage_is_deleted and negative_condition props to
            # false when is_switch_mode=true so that when we have an event on a
            # negative_condition=true edge it will find the correct subgraph
            self._switch_edge_negative_props(is_switch_mode, scenario_edge,
                                             subgraph, False)

            initial_map = Mapping(scenario_edge.edge, edge, False)
            curr_matches = \
                self._entity_graph.algo.sub_graph_matching(subgraph,
                                                           initial_map)

            # switch back to the original values
            self._switch_edge_negative_props(is_switch_mode, scenario_edge,
                                             subgraph, True)

            self._remove_negative_vertices_from_matches(curr_matches,
                                                        connected_component)

            matches.append((is_switch_mode, curr_matches))
        return matches

    def get_connected_component(self, subgraph, target):
        connected_component = self.connected_component_cache.get(
            id(subgraph), {}).get(id(target))
        if not connected_component:
            connected_component = subgraph.algo.graph_query_vertices(
                root_id=target,
                edge_query_dict={'!=': {NEG_CONDITION: True}})
            self.connected_component_cache[id(subgraph)][id(target)] = \
                connected_component
        return connected_component

    def _db_action_to_action_info(self, db_action):
        target = self._entity_graph.get_vertex(db_action.target_vertex_id)
        targets = {TARGET: target}
        if db_action.source_vertex_id:
            source = self._entity_graph.get_vertex(db_action.source_vertex_id)
            targets[SOURCE] = source
        scenario_action = self._scenario_repo.actions.get(db_action.action_id)
        properties = copy.copy(scenario_action.properties)
        action_specs = ActionSpecs(
            id=db_action.action_id,
            type=db_action.action_type,
            targets=targets,
            properties=properties,
        )
        action_info = ActionInfo(
            specs=action_specs,
            mode=ActionMode.DO,
            action_id=db_action.action_id,
            trigger_id=db_action.trigger,
        )
        return action_info

    @staticmethod
    def _switch_edge_negative_props(is_switch_mode,
                                    scenario_edge,
                                    subgraph,
                                    status):
        if is_switch_mode:
            scenario_edge.edge[NEG_CONDITION] = status
            scenario_edge.edge[EProps.VITRAGE_IS_DELETED] = status
            subgraph.update_edge(scenario_edge.edge)

    @staticmethod
    def _remove_negative_vertices_from_matches(matches, connected_component):
        for match in matches:
            ver_ids = [v.vertex_id for v in connected_component.get_vertices()]
            ver_to_remove = [id for id in match.keys() if id not in ver_ids]
            for v_id in ver_to_remove:
                del match[v_id]


class ActiveActionsTracker(object):
    """Keeps track of all active actions and relative dominance/priority.

    Actions are organized according to resource-id and action details.

    Examples:

    - all set_state actions on a given resource are considered similar action
    regardless of state
    - all raise_alarm of type alarm_name on a given resource are considered
    similar action, regardless of severity

    Each action is assigned a score by mapping the value property to the
    priority defined in datasource values config.

    - Alarm: severity
    - Resource: state

    The score is used to determine which action in each group of similar
    actions to be executed next.
    """
    action_tools = None

    def __init__(self, db, actions):
        self.db = db
        self.data = defaultdict(set)
        self.actions_to_create = {}
        self.actions_to_remove = set()
        self.actions_to_perform = []  # use a list to keep the insertion order
        self._init_action_tools()

        # Query DB for all actions with same properties
        actions_keys = set([self._get_key(action) for action in actions])
        db_rows = self.db.active_actions.query_similar(actions_keys) or []
        for db_row in db_rows:
            self.data[(db_row.source_vertex_id, db_row.target_vertex_id,
                       db_row.extra_info, db_row.action_type)].add(db_row)

    @classmethod
    def _init_action_tools(cls):
        if cls.action_tools:
            return
        info_mapper = DatasourceInfoMapper()
        alarms_score = info_mapper.get_datasource_priorities('vitrage')
        all_scores = info_mapper.get_datasource_priorities()
        cls.action_tools = {
            ActionType.SET_STATE: pt.SetStateTools(all_scores),
            ActionType.RAISE_ALARM: pt.RaiseAlarmTools(alarms_score),
            ActionType.ADD_CAUSAL_RELATIONSHIP: pt.BaselineTools,
            ActionType.MARK_DOWN: pt.BaselineTools,
            ActionType.EXECUTE_MISTRAL: pt.BaselineTools
        }

    def calc_do_action(self, action_info):
        """Add this action to active_actions table, if not exists

        return value to help decide if action should be performed
        Only a top scored action that is new should be performed
        :return: (is top score, is it already existing)
        """
        similar_actions = self._get_similar(action_info)
        exists = any(
            a.action_id == action_info.action_id and
            a.trigger == action_info.trigger_id for a in similar_actions)
        if not exists:
            self._add(action_info)
        if not exists and self._is_highest_score(similar_actions, action_info):
            self.actions_to_perform.append(action_info)

    def calc_undo_action(self, action_info):
        """Delete this action form active_actions table, if exists

        decide if action should be performed
        A top scored action should be 'undone' if there is not a second action.
        If there is a second, it should now be 'done' and become the dominant
        :param action_info: action to delete
        """
        similar_actions = self._get_similar(action_info)
        if not self._is_highest_score(similar_actions, action_info):
            self._remove(action_info)
            return

        second_highest = self._sort_db_actions(similar_actions)[1]\
            if len(similar_actions) > 1 else None
        # We should 'DO' the Second highest scored action so
        # to override the existing dominant action.
        # or, if there is no second highest scored action
        # So we just 'UNDO' the existing dominant action
        if second_highest:
            self.actions_to_perform.append(second_highest)
        else:
            self.actions_to_perform.append(action_info)
        self._remove(action_info)

    def flush_db_updates(self):
        self.db.active_actions.bulk_create(self.actions_to_create.values())
        self.db.active_actions.bulk_delete(self.actions_to_remove)

    def _add(self, action_info):
        db_row = self._to_db_row(action_info)
        self._get_similar(action_info).add(db_row)
        id_ = ScenarioEvaluator._generate_action_id(action_info.specs)
        if id_ not in self.actions_to_create:
            self.actions_to_create[id_] = db_row

    def _remove(self, action_info):
        similar_actions = self._get_similar(action_info)
        for action in similar_actions:
            if action.trigger == action_info.trigger_id and \
                    action.action_id == action_info.action_id:
                similar_actions.remove(action)
                break
        self.actions_to_remove.add(
            (action_info.trigger_id, action_info.action_id))

    def _get_similar(self, action_info):
        return self.data.get(self._get_key(action_info), set())

    def _get_key(self, action_info):
        src = action_info.specs.targets.get(SOURCE, {}).get(VProps.VITRAGE_ID)
        trg = action_info.specs.targets.get(TARGET, {}).get(VProps.VITRAGE_ID)
        extra_info = self.action_tools[action_info.specs.type].get_extra_info(
            action_info.specs)
        action_type = action_info.specs.type
        return src, trg, extra_info, action_type

    def _to_db_row(self, action_info):
        source = action_info.specs.targets.get(SOURCE, {})
        target = action_info.specs.targets.get(TARGET, {})
        action_score = self.action_tools[action_info.specs.type]. \
            get_score(action_info)
        extra_info = self.action_tools[action_info.specs.type]. \
            get_extra_info(action_info.specs)
        return storage.sqlalchemy.models.ActiveAction(
            action_type=action_info.specs.type,
            extra_info=extra_info,
            source_vertex_id=source.get(VProps.VITRAGE_ID),
            target_vertex_id=target.get(VProps.VITRAGE_ID),
            action_id=action_info.action_id,
            trigger=action_info.trigger_id,
            score=action_score)

    @classmethod
    def _is_highest_score(cls, db_actions, action_info):
        """Get the top action from the list and compare to action_info

        Actions are sorted according to:
        score - primary, ascending
        created_at - secondary, descending
        """
        if not db_actions:
            return True
        highest_score_action = min(
            db_actions, key=lambda action: (-action.score, action.created_at
                                            or utcnow(False)))
        return highest_score_action.trigger == action_info.trigger_id and \
            highest_score_action.action_id == action_info.action_id

    @staticmethod
    def _sort_db_actions(db_actions):
        """Sort ActiveAction items by two fields

        score - primary, ascending
        created_at - secondary, descending
        """
        return sorted(
            db_actions,
            key=lambda action: (-action.score, action.created_at),
            reverse=False)



# Copyright 2018 - Nokia Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_log import log
from oslo_utils.strutils import bool_from_string
import pecan
from pecan.core import abort

from vitrage.api.controllers.rest import RootRestController
from vitrage.api.policy import enforce
from vitrage.common.constants import TenantProps
from vitrage.common.constants import VertexProperties as Vprops
from vitrage.common.utils import decompress_obj

LOG = log.getLogger(__name__)


# noinspection PyBroadException
class BaseAlarmsController(RootRestController):

    @staticmethod
    def _get_alarms(**kwargs):
        vitrage_id = kwargs.get(Vprops.VITRAGE_ID)
        start = kwargs.get('start')
        end = kwargs.get('end')
        limit = kwargs.get('limit', 10000)
        sort_by = kwargs.get('sort_by', ['start_timestamp', 'vitrage_id'])
        sort_dirs = kwargs.get('sort_dirs', ['asc', 'asc'])
        filter_by = kwargs.get('filter_by', [])
        filter_vals = kwargs.get('filter_vals', [])
        next_page = kwargs.get('next_page', True)
        marker = kwargs.get('marker')
        only_active_alarms = kwargs.get('only_active_alarms', False)
        all_tenants = kwargs.get(TenantProps.ALL_TENANTS, False)
        all_tenants = bool_from_string(all_tenants)
        if all_tenants:
            enforce("list alarms:all_tenants", pecan.request.headers,
                    pecan.request.enforcer, {})
        else:
            enforce("list alarms", pecan.request.headers,
                    pecan.request.enforcer, {})

        alarms = \
            pecan.request.client.call(pecan.request.context,
                                      'get_alarms',
                                      vitrage_id=vitrage_id,
                                      all_tenants=all_tenants,
                                      start=start,
                                      end=end,
                                      limit=limit,
                                      sort_by=sort_by,
                                      sort_dirs=sort_dirs,
                                      filter_by=filter_by,
                                      filter_vals=filter_vals,
                                      next_page=next_page,
                                      marker=marker,
                                      only_active_alarms=only_active_alarms
                                      )

        try:
            alarms_list = decompress_obj(alarms)['alarms']
            return alarms_list

        except Exception:
            LOG.exception('Failed to get alarms')
            abort(404, 'Failed to get alarms')



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from vitrage.evaluator.template_fields import TemplateFields


def find_def_template(name, def_templates):

    for def_template_obj in def_templates.values():
        def_template = def_template_obj.data
        if def_template[TemplateFields.METADATA][TemplateFields.NAME] == name:

            return def_template
    return None



# Copyright 2016 - ZTE, Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import UpdateMethod
from vitrage.datasources.aodh import AODH_DATASOURCE
from vitrage.datasources.aodh.properties import AodhEventType
from vitrage.datasources.aodh.properties import AodhExtendedAlarmType as AType
from vitrage.datasources.aodh.properties import AodhProperties as AodhProps
from vitrage.tests import base
from vitrage.tests.mocks import mock_driver
from vitrage.tests.unit.datasources.aodh.mock_driver import MockAodhDriver


class AodhDriverTest(base.BaseTest):

    OPTS = [
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PUSH),
    ]

    # noinspection PyPep8Naming
    @classmethod
    def setUpClass(cls):
        super(AodhDriverTest, cls).setUpClass()
        cls.conf = cfg.ConfigOpts()
        cls.conf.register_opts(cls.OPTS, group=AODH_DATASOURCE)

    def test_event_alarm_notifications(self):

        aodh_driver = MockAodhDriver()

        # 1. alarm creation with 'ok' state
        # prepare data
        detail_data = {"type": "creation",
                       AodhProps.DETAIL: self._create_alarm_data_type_event()}
        generators = \
            mock_driver.simple_aodh_alarm_notification_generators(
                alarm_num=1,
                update_events=1,
                update_vals=detail_data)
        alarm = mock_driver.generate_sequential_events_list(generators)[0]
        alarm_info = alarm.copy()

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.CREATION)

        # Test assertions
        # alarm with status OK should not be handled
        self.assertIsNone(entity)

        # 2.alarm state transition from 'ok' to 'alarm'
        detail_data = {"type": "state transition",
                       AodhProps.DETAIL: {AodhProps.STATE: "alarm"}}
        alarm.update(detail_data)
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.STATE_TRANSITION)

        # Test assertions
        # alarm state change: ok->alarm, need to be added
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.STATE],
                         alarm[AodhProps.DETAIL][AodhProps.STATE])
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.SEVERITY])
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.STATE_TRANSITION)

        # 3. delete alarm which is 'alarm' state
        # prepare data
        detail_data = {"type": "deletion"}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.DELETION)

        # Test assertions
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.DELETION)

        # 4. alarm creation with 'alarm' state
        # prepare data
        detail_data = {"type": "creation",
                       AodhProps.DETAIL:
                           self._create_alarm_data_type_event(state="alarm")}
        generators = \
            mock_driver.simple_aodh_alarm_notification_generators(
                alarm_num=1,
                update_events=1,
                update_vals=detail_data)
        alarm = mock_driver.generate_sequential_events_list(generators)[0]
        alarm_info = alarm.copy()

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.CREATION)

        # Test assertions
        # alarm with status 'alarm' need to be added
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.STATE],
                         alarm[AodhProps.DETAIL][AodhProps.STATE])
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.SEVERITY])
        self.assertIsNone(entity[AodhProps.RESOURCE_ID])
        self.assertEqual("*", entity[AodhProps.EVENT_TYPE])
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.CREATION)

        # 5. alarm rule change
        # prepare data
        detail_data = {"type": "rule change",
                       AodhProps.DETAIL: {
                           "severity": "critical",
                           AodhProps.RULE:
                               {"query": [{"field": "traits.resource_id",
                                           "type": "",
                                           "value": "1",
                                           "op": "eq"}],
                                "event_type": "instance.update"}}}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.RULE_CHANGE)

        # Test assertions
        # alarm rule change: need to be update
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.DETAIL][AodhProps.SEVERITY])
        self.assertEqual(
            entity[AodhProps.EVENT_TYPE],
            alarm[AodhProps.DETAIL][AodhProps.RULE][AodhProps.EVENT_TYPE])
        self.assertEqual("1", entity[AodhProps.RESOURCE_ID])
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.RULE_CHANGE)

        # 6. alarm state change from 'alarm' to 'ok'
        # prepare data
        detail_data = {"type": "state transition",
                       AodhProps.DETAIL: {AodhProps.STATE: "ok"}}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.STATE_TRANSITION)

        # Test assertions
        # alarm state change: alarm->OK, need to be deleted
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.STATE_TRANSITION)

        # 7. delete alarm which is 'ok' state
        # prepare data
        detail_data = {"type": "deletion"}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.DELETION)

        # Test assertions
        self.assertIsNone(entity)

    def test_gnocchi_threshold_alarm_notifications(self):
        aodh_driver = MockAodhDriver()

        # 1. alarm creation with 'ok' state
        # prepare data
        detail_data = {"type": "gnocchi_resources_threshold",
                       AodhProps.DETAIL: self._create_alarm_data_gnocchi()}
        generators = \
            mock_driver.simple_aodh_alarm_notification_generators(
                alarm_num=1,
                update_events=1,
                update_vals=detail_data)
        alarm = mock_driver.generate_sequential_events_list(generators)[0]
        alarm_info = alarm.copy()

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.CREATION)

        # Test assertions
        # alarm with status OK should not be handled
        self.assertIsNone(entity)

        # 2.alarm state transition from 'ok' to 'alarm'
        detail_data = {"type": "state transition",
                       AodhProps.DETAIL: {AodhProps.STATE: "alarm"}}
        alarm.update(detail_data)
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.STATE_TRANSITION)

        # Test assertions
        # alarm state change: ok->alarm, need to be added
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.STATE],
                         alarm[AodhProps.DETAIL][AodhProps.STATE])
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.SEVERITY])

        # 3. delete alarm which is 'alarm' state
        # prepare data
        detail_data = {"type": "deletion"}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.DELETION)

        # Test assertions
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.DELETION)

        # 4. alarm creation with 'alarm' state
        # prepare data
        detail_data = {"type": "gnocchi_resources_threshold",
                       AodhProps.DETAIL:
                           self._create_alarm_data_gnocchi(state="alarm")}
        generators = \
            mock_driver.simple_aodh_alarm_notification_generators(
                alarm_num=1,
                update_events=1,
                update_vals=detail_data)
        alarm = mock_driver.generate_sequential_events_list(generators)[0]
        alarm_info = alarm.copy()

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.CREATION)

        # Test assertions
        # alarm with status 'alarm' need to be added
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.STATE],
                         alarm[AodhProps.DETAIL][AodhProps.STATE])
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.SEVERITY])
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.CREATION)

        # 5. alarm rule change
        # prepare data
        detail_data = {"type": "rule change",
                       AodhProps.DETAIL: {
                           "severity": "critical",
                           AodhProps.RULE:
                               {"granularity": "300",
                                "threshold": "0.0123",
                                "comparison_operator": "eq"}}}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.RULE_CHANGE)

        # Test assertions
        # alarm rule change: need to be update
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.DETAIL][AodhProps.SEVERITY])
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.RULE_CHANGE)

        # 6. alarm state change from 'alarm' to 'ok'
        # prepare data
        detail_data = {"type": "state transition",
                       AodhProps.DETAIL: {AodhProps.STATE: "ok"}}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.STATE_TRANSITION)

        # Test assertions
        # alarm state change: alarm->OK, need to be deleted
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.STATE_TRANSITION)

        # 7. delete alarm which is 'ok' state
        # prepare data
        detail_data = {"type": "deletion"}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.DELETION)

        # Test assertions
        self.assertIsNone(entity)

    def test_gnocchi_aggregation_by_metrics_alarm_notifications(self):
        aodh_driver = MockAodhDriver()

        # 1. alarm creation with 'ok' state
        # prepare data
        detail_data = {
            "type": AType.GNOCCHI_AGGREGATION_BY_METRICS_THRESHOLD,
            AodhProps.DETAIL: self._create_alarm_data_metrics()
        }
        generators = \
            mock_driver.simple_aodh_alarm_notification_generators(
                alarm_num=1,
                update_events=1,
                update_vals=detail_data)
        alarm = mock_driver.generate_sequential_events_list(generators)[0]
        alarm_info = alarm.copy()

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.CREATION)

        # Test assertions
        # alarm with status OK should not be handled
        self.assertIsNone(entity)

        # 2.alarm state transition from 'ok' to 'alarm'
        detail_data = {"type": "state transition",
                       AodhProps.DETAIL: {AodhProps.STATE: "alarm"}}
        alarm.update(detail_data)
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.STATE_TRANSITION)

        # Test assertions
        # alarm state change: ok->alarm, need to be added
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.STATE],
                         alarm[AodhProps.DETAIL][AodhProps.STATE])
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.SEVERITY])

        # 3. delete alarm which is 'alarm' state
        # prepare data
        detail_data = {"type": "deletion"}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.DELETION)

        # Test assertions
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.DELETION)

        # 4. alarm creation with 'alarm' state
        # prepare data
        detail_data = {
            "type": AType.GNOCCHI_AGGREGATION_BY_METRICS_THRESHOLD,
            AodhProps.DETAIL: self._create_alarm_data_metrics(state="alarm")
        }
        generators = \
            mock_driver.simple_aodh_alarm_notification_generators(
                alarm_num=1,
                update_events=1,
                update_vals=detail_data)
        alarm = mock_driver.generate_sequential_events_list(generators)[0]
        alarm_info = alarm.copy()

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.CREATION)

        # Test assertions
        # alarm with status 'alarm' need to be added
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.STATE],
                         alarm[AodhProps.DETAIL][AodhProps.STATE])
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.SEVERITY])
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.CREATION)

        # 5. alarm rule change
        # prepare data
        detail_data = {"type": "rule change",
                       AodhProps.DETAIL: {
                           "severity": "critical",
                           AodhProps.RULE:
                               {"granularity": "300",
                                "threshold": "0.0123",
                                "comparison_operator": "eq"}}}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.RULE_CHANGE)

        # Test assertions
        # alarm rule change: need to be update
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.DETAIL][AodhProps.SEVERITY])
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.RULE_CHANGE)

        # 6. alarm state change from 'alarm' to 'ok'
        # prepare data
        detail_data = {"type": "state transition",
                       AodhProps.DETAIL: {AodhProps.STATE: "ok"}}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.STATE_TRANSITION)

        # Test assertions
        # alarm state change: alarm->OK, need to be deleted
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.STATE_TRANSITION)

        # 7. delete alarm which is 'ok' state
        # prepare data
        detail_data = {"type": "deletion"}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.DELETION)

        # Test assertions
        self.assertIsNone(entity)

    def test_gnocchi_aggregation_by_resource_alarm_notifications(self):
        aodh_driver = MockAodhDriver()

        # 1. alarm creation with 'ok' state
        # prepare data
        detail_data = {
            "type": AType.GNOCCHI_AGGREGATION_BY_RESOURCES_THRESHOLD,
            AodhProps.DETAIL: self._create_alarm_data_resource()
        }
        generators = \
            mock_driver.simple_aodh_alarm_notification_generators(
                alarm_num=1,
                update_events=1,
                update_vals=detail_data)
        alarm = mock_driver.generate_sequential_events_list(generators)[0]
        alarm_info = alarm.copy()

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.CREATION)

        # Test assertions
        # alarm with status OK should not be handled
        self.assertIsNone(entity)

        # 2.alarm state transition from 'ok' to 'alarm'
        detail_data = {"type": "state transition",
                       AodhProps.DETAIL: {AodhProps.STATE: "alarm"}}
        alarm.update(detail_data)
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.STATE_TRANSITION)

        # Test assertions
        # alarm state change: ok->alarm, need to be added
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.STATE],
                         alarm[AodhProps.DETAIL][AodhProps.STATE])
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.SEVERITY])

        # 3. delete alarm which is 'alarm' state
        # prepare data
        detail_data = {"type": "deletion"}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.DELETION)

        # Test assertions
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.DELETION)

        # 4. alarm creation with 'alarm' state
        # prepare data
        detail_data = {
            "type": AType.GNOCCHI_AGGREGATION_BY_RESOURCES_THRESHOLD,
            AodhProps.DETAIL:
            self._create_alarm_data_gnocchi(state="alarm")
        }
        generators = \
            mock_driver.simple_aodh_alarm_notification_generators(
                alarm_num=1,
                update_events=1,
                update_vals=detail_data)
        alarm = mock_driver.generate_sequential_events_list(generators)[0]
        alarm_info = alarm.copy()

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.CREATION)

        # Test assertions
        # alarm with status 'alarm' need to be added
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.STATE],
                         alarm[AodhProps.DETAIL][AodhProps.STATE])
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.SEVERITY])
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.CREATION)

        # 5. alarm rule change
        # prepare data
        detail_data = {"type": "rule change",
                       AodhProps.DETAIL: {
                           "severity": "critical",
                           AodhProps.RULE:
                               {"granularity": "300",
                                "threshold": "0.0123",
                                "comparison_operator": "eq"}}}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.RULE_CHANGE)

        # Test assertions
        # alarm rule change: need to be update
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.DETAIL][AodhProps.SEVERITY])
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.RULE_CHANGE)

        # 6. alarm state change from 'alarm' to 'ok'
        # prepare data
        detail_data = {"type": "state transition",
                       AodhProps.DETAIL: {AodhProps.STATE: "ok"}}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.STATE_TRANSITION)

        # Test assertions
        # alarm state change: alarm->OK, need to be deleted
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.STATE_TRANSITION)

        # 7. delete alarm which is 'ok' state
        # prepare data
        detail_data = {"type": "deletion"}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.DELETION)

        # Test assertions
        self.assertIsNone(entity)

    def test_composite_alarm_notifications(self):
        aodh_driver = MockAodhDriver()

        # 1. alarm creation with 'ok' state
        # prepare data
        detail_data = {"type": "composite",
                       AodhProps.DETAIL: self._create_alarm_data_composite()}
        generators = \
            mock_driver.simple_aodh_alarm_notification_generators(
                alarm_num=1,
                update_events=1,
                update_vals=detail_data)
        alarm = mock_driver.generate_sequential_events_list(generators)[0]
        alarm_info = alarm.copy()

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.CREATION)

        # Test assertions
        # alarm with status OK should not be handled
        self.assertIsNone(entity)

        # 2.alarm state transition from 'ok' to 'alarm'
        detail_data = {"type": "state transition",
                       AodhProps.DETAIL: {AodhProps.STATE: "alarm"}}
        alarm.update(detail_data)
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.STATE_TRANSITION)

        # Test assertions
        # alarm state change: ok->alarm, need to be added
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.STATE],
                         alarm[AodhProps.DETAIL][AodhProps.STATE])
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.SEVERITY])

        # 3. delete alarm which is 'alarm' state
        # prepare data
        detail_data = {"type": "deletion"}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.DELETION)

        # Test assertions
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.DELETION)

        # 4. alarm creation with 'alarm' state
        # prepare data
        detail_data = {"type": "composite",
                       AodhProps.DETAIL:
                           self._create_alarm_data_composite(state="alarm")}
        generators = \
            mock_driver.simple_aodh_alarm_notification_generators(
                alarm_num=1,
                update_events=1,
                update_vals=detail_data)
        alarm = mock_driver.generate_sequential_events_list(generators)[0]
        alarm_info = alarm.copy()

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.CREATION)

        # Test assertions
        # alarm with status 'alarm' need to be added
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.STATE],
                         alarm[AodhProps.DETAIL][AodhProps.STATE])
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.SEVERITY])
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.CREATION)

        # 5. alarm rule change
        # prepare data
        detail_data = {"type": "rule change",
                       AodhProps.DETAIL: {
                           "severity": "critical",
                           AodhProps.RULE:
                               {"granularity": "300",
                                "threshold": "0.0123",
                                "comparison_operator": "eq"}}}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.RULE_CHANGE)

        # Test assertions
        # alarm rule change: need to be update
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[AodhProps.SEVERITY],
                         alarm[AodhProps.DETAIL][AodhProps.SEVERITY])
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.RULE_CHANGE)

        # 6. alarm state change from 'alarm' to 'ok'
        # prepare data
        detail_data = {"type": "state transition",
                       AodhProps.DETAIL: {AodhProps.STATE: "ok"}}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm,
                                          AodhEventType.STATE_TRANSITION)

        # Test assertions
        # alarm state change: alarm->OK, need to be deleted
        self.assertIsNotNone(entity)
        self._validate_aodh_entity_comm_props(entity, alarm_info)
        self.assertEqual(entity[DSProps.EVENT_TYPE],
                         AodhEventType.STATE_TRANSITION)

        # 7. delete alarm which is 'ok' state
        # prepare data
        detail_data = {"type": "deletion"}
        alarm.update(detail_data)

        # action
        entity = aodh_driver.enrich_event(alarm, AodhEventType.DELETION)

        # Test assertions
        self.assertIsNone(entity)

    def _create_alarm_data_composite(self,
                                     state='ok',
                                     type='composite',
                                     rule=None):
        if rule is None:
            rule = {"or":
                    [{"evaluation_periods": 1,
                      "metrics": ["6ade05e5-f98b-4b7d-a0b3-9d330c4c3c41"],
                      "aggregation_method": "mean",
                      "granularity": 60,
                      "threshold": 100.0,
                      "type": "gnocchi_aggregation_by_metrics_threshold",
                      "comparison_operator": "lt"},
                     {"evaluation_periods": 3,
                      "metrics": ["89vde0e5-k3rb-4b7d-a0b3-9d330c4c3c41"],
                      "aggregation_method": "mean",
                      "granularity": 2,
                      "threshold": 80.0,
                      "type": "gnocchi_aggregation_by_metrics_threshold",
                      "comparison_operator": "ge"}
                     ]}
        return {AodhProps.DESCRIPTION: "test",
                AodhProps.TIMESTAMP: "2016-11-09T01:39:13.839584",
                AodhProps.ENABLED: True,
                AodhProps.STATE_TIMESTAMP: "2016-11-09T01:39:13.839584",
                AodhProps.ALARM_ID: "7e5c3754-e2eb-4782-ae00-7da5ded8568b",
                AodhProps.REPEAT_ACTIONS: False,
                AodhProps.PROJECT_ID: "c365d18fcc03493187016ae743f0cc4d",
                AodhProps.NAME: "test",
                AodhProps.SEVERITY: "low",
                AodhProps.RESOURCE_ID: "88cd2d1d-8af4-4d00-9b5e-f82f8c8b0f8d",
                AodhProps.TYPE: type,
                AodhProps.STATE: state,
                AodhProps.RULE: rule}

    def _create_alarm_data_metrics(
            self,
            state='ok',
            type=AType.GNOCCHI_AGGREGATION_BY_METRICS_THRESHOLD,
            rule=None
    ):
        if rule is None:
            rule = {"threshold": '100',
                    "aggregation_method": "mean",
                    "comparison_operator": "lt"
                    }

        return {AodhProps.DESCRIPTION: "metric test",
                AodhProps.TIMESTAMP: "2017-04-03T01:39:13.839584",
                AodhProps.ENABLED: True,
                AodhProps.STATE_TIMESTAMP: "2017-04-03T01:39:13.839584",
                AodhProps.ALARM_ID: "7e5c3754-e2eb-4782-ae00-7da5ded8568b",
                AodhProps.REPEAT_ACTIONS: False,
                AodhProps.PROJECT_ID: "c365d18fcc03493187016ae743f0cc4d",
                AodhProps.NAME: "test",
                AodhProps.SEVERITY: "low",
                AodhProps.RESOURCE_ID: "88cd2d1d-8af4-4d00-9b5e-f82f8c8b0f8d",
                AodhProps.METRICS: "6ade05e5-f98b-4b7d-a0b3-9d330c4c3c41",
                AodhProps.TYPE: type,
                AodhProps.STATE: state,
                AodhProps.RULE: rule}

    def _create_alarm_data_resource(
            self,
            state='ok',
            type=AType.GNOCCHI_AGGREGATION_BY_RESOURCES_THRESHOLD,
            rule=None):
        if rule is None:
            rule = {"evaluation_periods": 3,
                    "metric": "cpu_util",
                    "aggregation_method": "mean",
                    "granularity": 300,
                    "threshold": 50.0,
                    "query": [{"=":
                              {"resource_id":
                               "6df1747a-ef31-4897-854e-ffa2ae568e45"}}],
                    "comparison_operator": "ge",
                    "resource_type": "instance"
                    }

        return {AodhProps.DESCRIPTION: "test",
                AodhProps.TIMESTAMP: "2016-11-09T01:39:13.839584",
                AodhProps.ENABLED: True,
                AodhProps.STATE_TIMESTAMP: "2016-11-09T01:39:13.839584",
                AodhProps.ALARM_ID: "7e5c3754-e2eb-4782-ae00-7da5ded8568b",
                AodhProps.REPEAT_ACTIONS: False,
                AodhProps.PROJECT_ID: "c365d18fcc03493187016ae743f0cc4d",
                AodhProps.NAME: "test",
                AodhProps.SEVERITY: "low",
                AodhProps.RESOURCE_ID: "88cd2d1d-8af4-4d00-9b5e-f82f8c8b0f8d",
                AodhProps.TYPE: type,
                AodhProps.STATE: state,
                AodhProps.RULE: rule}

    def _create_alarm_data_gnocchi(self,
                                   state="ok",
                                   type="gnocchi_resources_threshold",
                                   rule=None):

        if rule is None:
            rule = {"granularity": "300",
                    "threshold": "0.001",
                    "comparison_operator": "gt",
                    "resource_type": "instance",
                    AodhProps.RESOURCE_ID:
                        "88cd2d1d-8af4-4d00-9b5e-f82f8c8b0f8d"
                    }
        return {AodhProps.DESCRIPTION: "test",
                AodhProps.TIMESTAMP: "2016-11-09T01:39:13.839584",
                AodhProps.ENABLED: True,
                AodhProps.STATE_TIMESTAMP: "2016-11-09T01:39:13.839584",
                AodhProps.ALARM_ID: "7e5c3754-e2eb-4782-ae00-7da5ded8568b",
                AodhProps.REPEAT_ACTIONS: False,
                AodhProps.PROJECT_ID: "c365d18fcc03493187016ae743f0cc4d",
                AodhProps.NAME: "test",
                AodhProps.SEVERITY: "low",
                AodhProps.RESOURCE_ID: "88cd2d1d-8af4-4d00-9b5e-f82f8c8b0f8d",
                AodhProps.TYPE: type,
                AodhProps.STATE: state,
                AodhProps.RULE: rule}

    def _create_alarm_data_type_event(self,
                                      state="ok",
                                      type="event",
                                      rule=None):

        if rule is None:
            rule = {"query": [], "event_type": "*"}
        return {AodhProps.DESCRIPTION: "test",
                AodhProps.TIMESTAMP: "2016-11-09T01:39:13.839584",
                AodhProps.ENABLED: True,
                AodhProps.STATE_TIMESTAMP: "2016-11-09T01:39:13.839584",
                AodhProps.ALARM_ID: "7e5c3754-e2eb-4782-ae00-7da5ded8568b",
                AodhProps.REPEAT_ACTIONS: False,
                AodhProps.PROJECT_ID: "c365d18fcc03493187016ae743f0cc4d",
                AodhProps.NAME: "test",
                AodhProps.SEVERITY: "low",
                AodhProps.TYPE: type,
                AodhProps.STATE: state,
                AodhProps.RULE: rule}

    def _validate_aodh_entity_comm_props(self, entity, alarm):

        self.assertEqual(entity[AodhProps.ALARM_ID],
                         alarm[AodhProps.ALARM_ID])
        self.assertEqual(entity[AodhProps.PROJECT_ID],
                         alarm[AodhProps.PROJECT_ID])
        self.assertEqual(entity[AodhProps.TIMESTAMP],
                         alarm[AodhProps.TIMESTAMP])
        self.assertEqual(entity[AodhProps.DESCRIPTION],
                         alarm[AodhProps.DETAIL][AodhProps.DESCRIPTION])
        self.assertEqual(entity[AodhProps.ENABLED],
                         alarm[AodhProps.DETAIL][AodhProps.ENABLED])
        self.assertEqual(entity[AodhProps.NAME],
                         alarm[AodhProps.DETAIL][AodhProps.NAME])
        self.assertEqual(entity[AodhProps.REPEAT_ACTIONS],
                         alarm[AodhProps.DETAIL][AodhProps.REPEAT_ACTIONS])
        self.assertEqual(entity[AodhProps.TYPE],
                         alarm[AodhProps.DETAIL][AodhProps.TYPE])



# Copyright 2016 - Nokia, ZTE
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR  CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import UpdateMethod

STATIC_DATASOURCE = 'static'

OPTS = [
    cfg.StrOpt(DSOpts.TRANSFORMER,
               default='vitrage.datasources.static.transformer.'
                       'StaticTransformer',
               help='Static transformer class path',
               required=True),
    cfg.StrOpt(DSOpts.DRIVER,
               default='vitrage.datasources.static.driver.'
                       'StaticDriver',
               help='Static driver class path',
               required=True),
    cfg.StrOpt(DSOpts.UPDATE_METHOD,
               default=UpdateMethod.PULL,
               help='None: updates only via Vitrage periodic snapshots.'
                    'Pull: updates periodically.'
                    'Push: updates by getting notifications from the'
                    ' datasource itself.',
               required=True),
    cfg.IntOpt(DSOpts.CHANGES_INTERVAL,
               default=30,
               min=10,
               help='interval in seconds between checking changes in the'
                    'static configuration files'),
    cfg.StrOpt('directory', default='/etc/vitrage/static_datasources',
               help='static data sources configuration directory')]


STATIC_SCHEMA = {
    "type": "object",
    "required": ["definitions"],
    "properties": {
        "metadata": {
            "type": "object",
            "required": ["name"],
            "properties": {
                "name": {
                    "type": "string",
                },
                "description": {
                    "type": "string",
                },
            }
        },
        "definitions": {
            "type": "object",
            "required": ["entities", "relationships"],
            "properties": {
                "entities": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "required": ["static_id"],
                        "properties": {
                            "static_id": {
                                "type": "string",
                            },
                            "type": {
                                "type": "string",
                            },
                            "name": {
                                "type": "string",
                            },
                            "id": {
                                "type": "string",
                            },
                            "state": {
                                "type": "string",
                            },
                        },
                    },
                },
                "relationships": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "required": ["source", "target",
                                     "relationship_type"],
                        "properties": {
                            "source": {
                                "type": "string",
                            },
                            "target": {
                                "type": "string",
                            },
                            "relationship_type": {
                                "type": "string",
                            },
                        },
                    },
                },
            }
        },
    },
}


class StaticFields(object):
    """yaml fields for static definitions"""
    METADATA = 'metadata'
    DEFINITIONS = 'definitions'
    RELATIONSHIPS = 'relationships'
    ENTITIES = 'entities'
    RELATIONSHIP_TYPE = 'relationship_type'
    SOURCE = 'source'
    TARGET = 'target'
    STATIC_ID = 'static_id'
    TYPE = 'type'
    CATEGORY = 'category'
    ID = 'id'
    NAME = 'name'
    STATE = 'state'



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime

from oslo_config import cfg
from oslo_log import log as logging
from testtools import matchers

from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import EdgeLabel
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import UpdateMethod
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.cinder.volume import CINDER_VOLUME_DATASOURCE
from vitrage.datasources.cinder.volume.properties \
    import CinderProperties as CinderProps
from vitrage.datasources.cinder.volume.transformer \
    import CinderVolumeTransformer
from vitrage.datasources.nova.instance import NOVA_INSTANCE_DATASOURCE
from vitrage.datasources.nova.instance.transformer import InstanceTransformer
from vitrage.datasources import transformer_base as tbase
from vitrage.datasources.transformer_base import TransformerBase
from vitrage.tests import base
from vitrage.tests.mocks import mock_driver as mock_sync

LOG = logging.getLogger(__name__)


class TestCinderVolumeTransformer(base.BaseTest):

    OPTS = [
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PUSH),
    ]

    # noinspection PyAttributeOutsideInit,PyPep8Naming
    @classmethod
    def setUpClass(cls):
        super(TestCinderVolumeTransformer, cls).setUpClass()
        cls.transformers = {}
        cls.conf = cfg.ConfigOpts()
        cls.conf.register_opts(cls.OPTS, group=CINDER_VOLUME_DATASOURCE)
        cls.transformers[CINDER_VOLUME_DATASOURCE] = \
            CinderVolumeTransformer(cls.transformers)
        cls.transformers[NOVA_INSTANCE_DATASOURCE] = \
            InstanceTransformer(cls.transformers)

    def test_create_placeholder_vertex(self):
        LOG.debug('Cinder Volume transformer test: Create placeholder '
                  'vertex')

        # Tests setup
        volume_id = 'Instance123'
        timestamp = datetime.datetime.utcnow()
        properties = {
            VProps.ID: volume_id,
            VProps.VITRAGE_TYPE: CINDER_VOLUME_DATASOURCE,
            VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
            VProps.VITRAGE_SAMPLE_TIMESTAMP: timestamp
        }
        transformer = self.transformers[CINDER_VOLUME_DATASOURCE]

        # Test action
        placeholder = \
            transformer.create_neighbor_placeholder_vertex(**properties)

        # Test assertions)
        expected_key = \
            tbase.build_key(transformer._key_values(CINDER_VOLUME_DATASOURCE,
                                                    volume_id))
        expected_uuid = \
            TransformerBase.uuid_from_deprecated_vitrage_id(expected_key)
        self.assertEqual(expected_uuid, placeholder.vertex_id)

        observed_time = placeholder.get(VProps.VITRAGE_SAMPLE_TIMESTAMP)
        self.assertEqual(timestamp, observed_time)

        observed_type = placeholder.get(VProps.VITRAGE_TYPE)
        self.assertEqual(CINDER_VOLUME_DATASOURCE, observed_type)

        observed_entity_id = placeholder.get(VProps.ID)
        self.assertEqual(volume_id, observed_entity_id)

        observed_vitrage_category = placeholder.get(VProps.VITRAGE_CATEGORY)
        self.assertEqual(EntityCategory.RESOURCE, observed_vitrage_category)

        vitrage_is_placeholder = placeholder.get(VProps.VITRAGE_IS_PLACEHOLDER)
        self.assertTrue(vitrage_is_placeholder)

    def test_key_values(self):
        LOG.debug('Cinder Volume transformer test: get key values')

        # Test setup
        volume_type = CINDER_VOLUME_DATASOURCE
        volume_id = '12345'
        transformer = self.transformers[CINDER_VOLUME_DATASOURCE]

        # Test action
        observed_key_fields = transformer._key_values(volume_type,
                                                      volume_id)

        # Test assertions
        self.assertEqual(EntityCategory.RESOURCE, observed_key_fields[0])
        self.assertEqual(CINDER_VOLUME_DATASOURCE, observed_key_fields[1])
        self.assertEqual(volume_id, observed_key_fields[2])

    def test_snapshot_transform(self):
        LOG.debug('Cinder Volume transformer test: transform entity event '
                  'snapshot')

        # Test setup
        spec_list = mock_sync.simple_volume_generators(volume_num=3,
                                                       instance_num=7,
                                                       snapshot_events=7)
        static_events = mock_sync.generate_random_events_list(spec_list)

        for event in static_events:
            # Test action
            wrapper = self.transformers[CINDER_VOLUME_DATASOURCE].transform(
                event)

            # Test assertions
            vertex = wrapper.vertex
            self._validate_volume_vertex_props(vertex, event)

            neighbors = wrapper.neighbors
            self.assertThat(neighbors, matchers.HasLength(1))
            self._validate_neighbors(neighbors, vertex.vertex_id, event)

    def test_update_transform(self):
        LOG.debug('Cinder Volume transformer test: transform entity event '
                  'update')

        # Test setup
        spec_list = mock_sync.simple_volume_generators(volume_num=3,
                                                       instance_num=7,
                                                       update_events=7)
        static_events = mock_sync.generate_random_events_list(spec_list)

        for event in static_events:
            # Test action
            wrapper = self.transformers[CINDER_VOLUME_DATASOURCE].transform(
                event)

            # Test assertions
            vertex = wrapper.vertex
            self._validate_volume_vertex_props(vertex, event)

            neighbors = wrapper.neighbors
            self.assertThat(neighbors, matchers.HasLength(1))
            self._validate_neighbors(neighbors, vertex.vertex_id, event)

    def _validate_volume_vertex_props(self, vertex, event):

        is_update_event = tbase.is_update_event(event)

        self.assertEqual(EntityCategory.RESOURCE,
                         vertex[VProps.VITRAGE_CATEGORY])
        self.assertEqual(event[DSProps.ENTITY_TYPE],
                         vertex[VProps.VITRAGE_TYPE])

        id_field_path = 'volume_id' if is_update_event else 'id'
        self.assertEqual(
            tbase.extract_field_value(event, id_field_path),
            vertex[VProps.ID])

        self.assertEqual(event[DSProps.SAMPLE_DATE],
                         vertex[VProps.VITRAGE_SAMPLE_TIMESTAMP])

        name_field_path = 'display_name'
        self.assertEqual(
            tbase.extract_field_value(event, name_field_path),
            vertex[VProps.NAME])

        state_field_path = 'status'
        self.assertEqual(
            tbase.extract_field_value(event, state_field_path),
            vertex[VProps.STATE])

        size_field_path = 'size'
        self.assertEqual(
            tbase.extract_field_value(event, size_field_path),
            vertex[CinderProps.SIZE])

        volume_type_field_path = 'volume_type'
        self.assertEqual(
            tbase.extract_field_value(event, volume_type_field_path),
            vertex[CinderProps.VOLUME_TYPE])

        self.assertFalse(vertex[VProps.VITRAGE_IS_PLACEHOLDER])
        self.assertFalse(vertex[VProps.VITRAGE_IS_DELETED])

    def _validate_neighbors(self, neighbors, volume_vertex_id, event):
        instance_counter = 0

        for neighbor in neighbors:
            is_update_event = tbase.is_update_event(event)
            instance_id = event['volume_attachment'][0]['instance_uuid'] if \
                is_update_event else event['attachments'][0]['server_id']
            self._validate_instance_neighbor(neighbor,
                                             instance_id,
                                             volume_vertex_id)
            instance_counter += 1

        self.assertEqual(1,
                         instance_counter,
                         'Volume can be belonged to only one instance')

    def _validate_instance_neighbor(self,
                                    instance_neighbor,
                                    instance_id,
                                    volume_vertex_id):
        # validate neighbor vertex
        self.assertEqual(EntityCategory.RESOURCE,
                         instance_neighbor.vertex[VProps.VITRAGE_CATEGORY])
        self.assertEqual(NOVA_INSTANCE_DATASOURCE,
                         instance_neighbor.vertex[VProps.VITRAGE_TYPE])
        self.assertEqual(instance_id, instance_neighbor.vertex[VProps.ID])
        self.assertTrue(
            instance_neighbor.vertex[VProps.VITRAGE_IS_PLACEHOLDER])
        self.assertFalse(instance_neighbor.vertex[VProps.VITRAGE_IS_DELETED])

        # Validate neighbor edge
        edge = instance_neighbor.edge
        self.assertEqual(edge.target_id, instance_neighbor.vertex.vertex_id)
        self.assertEqual(edge.source_id, volume_vertex_id)
        self.assertEqual(edge.label, EdgeLabel.ATTACHED)



#  Copyright 2019 - Nokia Corporation
#  #
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#  #
#       http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from oslo_config import cfg

OPTS = [
    cfg.StrOpt('notifier',
               default='vitrage.notifier.plugins.zaqar.'
                       'zaqar_notifier.ZaqarNotifier',
               help='zaqar notifier class path',
               required=True),

    cfg.StrOpt('queue',
               default='alarms',
               help='zaqar queue to post messages',
               required=True),
]



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import namedtuple
from datetime import datetime

from oslo_config import cfg
from oslo_log import log

from vitrage.common.constants import NotifierEventTypes
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.alarm_properties import AlarmProperties as AlarmProps
from vitrage.datasources.transformer_base import TIMESTAMP_FORMAT
from vitrage.machine_learning.plugins.base import MachineLearningBase
import vitrage.machine_learning.plugins.jaccard_correlation.\
    accumulation_persistor_utils as APersistor
from vitrage.machine_learning.plugins.jaccard_correlation.\
    alarm_data_accumulator import AlarmDataAccumulator as ADAcummulator
from vitrage.machine_learning.plugins.jaccard_correlation.correlation_manager\
    import CorrelationManager as CM

CONF = cfg.CONF

LOG = log.getLogger(__name__)

AlarmID = namedtuple('AlarmID', [VProps.VITRAGE_RESOURCE_ID,
                                 VProps.VITRAGE_RESOURCE_TYPE,
                                 VProps.NAME])


class AlarmDataProcessor(MachineLearningBase):

    @staticmethod
    def get_plugin_name():
        return 'jaccard_correlation'

    def __init__(self):
        super(AlarmDataProcessor, self).__init__()
        self.data_manager = ADAcummulator(APersistor.load_data())
        self.correlation_manager = CM()
        self.num_of_events_to_flush = \
            CONF.jaccard_correlation.num_of_events_to_flush
        self.event_counter = 0

    def process_event(self, data, event_type):

        if event_type == NotifierEventTypes.ACTIVATE_ALARM_EVENT \
                or event_type == NotifierEventTypes.DEACTIVATE_ALARM_EVENT:

            # TODO(annarez): handle alarms from collectd
            if data[VProps.VITRAGE_TYPE] == 'collectd':
                return

            self._update_data_accumulator(data)
            self.event_counter += 1

            # flush all data once num_of_events_to_flush is achieved
            if self.event_counter == self.num_of_events_to_flush:
                LOG.debug("Persisting: %s", data)
                self.data_manager.flush_accumulations()
                APersistor.save_accumulated_data(self.data_manager)
                self.correlation_manager.output_correlations(self.data_manager)
                self.event_counter = 0

    def _update_data_accumulator(self, data):

        alarm_name = data[VProps.RAWTEXT] if data.get(VProps.RAWTEXT) else \
            data[VProps.NAME]

        alarm_id, timestamp = \
            self._get_alarm_id_and_timestamp(data, alarm_name)

        if data[VProps.STATE] == AlarmProps.ACTIVE_STATE:
            self.data_manager.append_active(alarm_id, timestamp)
        else:
            self.data_manager.append_inactive(alarm_id, timestamp)

    @staticmethod
    def _get_alarm_id_and_timestamp(data, alarm_name):

        alarm_id = AlarmID(data.get(VProps.VITRAGE_RESOURCE_ID),
                           data.get(VProps.VITRAGE_RESOURCE_TYPE),
                           alarm_name)

        timestamp = datetime.strptime(data[VProps.UPDATE_TIMESTAMP],
                                      TIMESTAMP_FORMAT)

        return alarm_id, timestamp



# Copyright 2020
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging

from vitrage.common.constants import EntityCategory
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.resource_transformer_base \
    import ResourceTransformerBase
from vitrage.datasources.tmfapi639 import TMFAPI639_DATASOURCE
from vitrage.datasources import transformer_base
import vitrage.graph.utils as graph_utils

from datetime import datetime

LOG = logging.getLogger(__name__)


class TmfApi639Transformer(ResourceTransformerBase):

    def __init__(self, transformers):
        super(TmfApi639Transformer, self).__init__(transformers)

    def _create_snapshot_entity_vertex(self, entity_event):
        return self._create_vertex(entity_event)

    def _create_update_entity_vertex(self, entity_event):
        return self._create_vertex(entity_event)

    def _create_snapshot_neighbors(self, entity_event):
        return self._create_tmfapi639_neighbors(entity_event)

    def _create_update_neighbors(self, entity_event):
        return self._create_tmfapi639_neighbors(entity_event)

    def _create_entity_key(self, entity_event):
        """the unique key of this entity"""
        entity_id = entity_event["id"]
        entity_type = TMFAPI639_DATASOURCE
        key_fields = self._key_values(entity_type, entity_id)
        return transformer_base.build_key(key_fields)

    def get_vitrage_type(self):
        return TMFAPI639_DATASOURCE

    def _create_vertex(self, entity_event):
        """Camps used from the received JSON:

        {id, name, @type ,resourceRelationship : [type, resource: id]}

        The TMF 639 API REST Endpoint can contain more information
        but we only use this one for topology.
        """
        sample_timestamp = \
            datetime.now().strftime(transformer_base.TIMESTAMP_FORMAT)
        update_timestamp = self._format_update_timestamp(
            update_timestamp=None,
            sample_timestamp=sample_timestamp)

        metadata = {
            VProps.NAME: entity_event["name"],
        }

        return graph_utils.create_vertex(
            self._create_entity_key(entity_event),
            vitrage_category=EntityCategory.RESOURCE,
            vitrage_type=TMFAPI639_DATASOURCE,
            vitrage_sample_timestamp=sample_timestamp,
            entity_id=entity_event["id"],
            update_timestamp=update_timestamp,
            entity_state='available',
            metadata=metadata)

    def _create_tmfapi639_neighbors(self, entity_event):
        neighbors_list = []
        for n in entity_event["resourceRelationship"]:
            # create placeholder vertex
            neigh = self._create_neighbor(
                entity_event,
                n["resource"]["id"],
                TMFAPI639_DATASOURCE,
                n["type"],
                is_entity_source=True)
            neighbors_list.append(neigh)
        return neighbors_list



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.common.constants import VertexProperties as VProps
from vitrage.tests import base


# noinspection PyProtectedMember
class BaseTransformerTest(base.BaseTest):

    def _validate_base_vertex_props(self,
                                    vertex,
                                    expected_name,
                                    expected_datasource_name):
        self.assertFalse(vertex[VProps.VITRAGE_IS_PLACEHOLDER])
        self.assertEqual(expected_datasource_name, vertex[VProps.VITRAGE_TYPE])
        self.assertEqual(expected_name, vertex[VProps.NAME])



# Copyright 2019 - China Mobile (SuZhou) Software Technology Co.,Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


class NetworkProperties(object):
    NAME = 'name'
    STATUS = 'status'
    ID = 'id'
    TENANT_ID = 'tenant_id'
    CREATED_AT = 'created_at'
    UPDATED_AT = 'updated_at'
    NETWORK = 'network'


class PortProperties(object):
    NAME = 'name'
    STATUS = 'status'
    ID = 'id'
    TENANT_ID = 'tenant_id'
    CREATED_AT = 'created_at'
    UPDATED_AT = 'updated_at'
    NETWORK = 'network'
    DEVICE_ID = 'device_id'
    NETWORK_ID = 'network_id'
    HOST_ID = 'host_id'
    PORT = 'port'
    IP_ADDRESS = 'ip_address'
    IP_ADDRESSES = 'ip_addresses'



# Copyright 2018 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# noinspection PyPackageRequirements
from unittest import mock

from oslo_config import cfg
# noinspection PyPackageRequirements
from testtools import matchers

from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import EventProperties as EProps
from vitrage.datasources.prometheus.driver import PROMETHEUS_EVENT_TYPE
from vitrage.datasources.prometheus.driver import PrometheusDriver
from vitrage.datasources.prometheus import PROMETHEUS_DATASOURCE
from vitrage.datasources.prometheus.properties \
    import PrometheusDatasourceProperties as PDProps
from vitrage.datasources.prometheus.properties \
    import PrometheusProperties as PProps
from vitrage.tests import base
from vitrage.tests.mocks import mock_driver
from vitrage.tests.mocks import utils


# noinspection PyProtectedMember
class PrometheusDriverTest(base.BaseTest):
    CONFIG_PATH = '/prometheus/prometheus_conf.yaml'
    OPTS = [
        cfg.StrOpt(
            DSOpts.CONFIG_FILE,
            help='Prometheus configuration file',
            default=utils.get_resources_dir() + CONFIG_PATH),
    ]

    def setUp(self):
        super(PrometheusDriverTest, self).setUp()
        self.conf_reregister_opts(self.OPTS, group=PROMETHEUS_DATASOURCE)

    @mock.patch('socket.gethostbyaddr')
    def test_adjust_label_value(self, mock_socket):

        # Test setup
        hostname = 'devstack-rocky-release-4'
        mock_socket.return_value = [hostname]
        driver = PrometheusDriver()
        valid_ip = '127.0.0.1'
        not_ip = 'localhost'
        invalid_ip = '127.1'

        # Test Action
        observed_valid_ip = driver._adjust_label_value(valid_ip)
        observed_not_ip = driver._adjust_label_value(not_ip)
        observed_invalid_ip = driver._adjust_label_value(invalid_ip)

        # Test assertions
        self.assertEqual(hostname, observed_valid_ip)
        self.assertEqual(not_ip, observed_not_ip)
        self.assertEqual(invalid_ip, observed_invalid_ip)

    @mock.patch('socket.gethostbyaddr')
    def test_calculate_host_vitrage_entity_unique_props(self, mock_socket):
        # First alert is 'AvgCPUTimeOnIdleMode' on host
        # Second alert is 'HighInodeUsage' on host

        # Test setup
        hostname = 'devstack-rocky-release-4'
        mock_socket.return_value = [hostname]
        driver = PrometheusDriver()
        alerts = self._generate_alerts()
        host_alert_1 = alerts[0]
        host_alert_2 = alerts[1]

        # Test Action
        observed_host_alert_1 = \
            driver._calculate_vitrage_entity_unique_props(host_alert_1)
        observed_host_alert_2 = \
            driver._calculate_vitrage_entity_unique_props(host_alert_2)

        # Test assertion
        expected = {'id': hostname}
        self.assertEqual(expected, observed_host_alert_1)
        self.assertEqual(expected, observed_host_alert_2)

    @mock.patch('socket.gethostbyaddr')
    def test_calculate_vm_vitrage_entity_unique_props(self, mock_socket):
        # First alert is 'HighTrafficOnBridge' on vm
        # Second alert is 'HighCpuOnVmAlert' on vm

        # Test setup
        hostname = 'devstack-rocky-release-4'
        mock_socket.return_value = [hostname]
        driver = PrometheusDriver()
        alerts = self._generate_alerts()
        vm_alert_1 = alerts[2]
        vm_alert_2 = alerts[3]

        # Test Action
        observed_vm_alert_1 = \
            driver._calculate_vitrage_entity_unique_props(vm_alert_1)
        observed_vm_alert_2 = \
            driver._calculate_vitrage_entity_unique_props(vm_alert_2)

        # Test assertion
        expected = {'instance_name': 'instance-00000005',
                    'host_id': hostname}
        self.assertEqual(expected, observed_vm_alert_1)
        self.assertEqual(expected, observed_vm_alert_2)

    def test_get_resource_alert_values(self):

        # Test setup
        driver = PrometheusDriver()
        alerts = self._generate_alerts()
        alert_1 = alerts[0]
        alert_2 = alerts[3]

        # Test Action
        observed_alert_1 = driver._get_resource_alert_values(alert_1)
        observed_alert_2 = driver._get_resource_alert_values(alert_2)

        # Test assertions
        expected_alert_1 = {'instance': '135.248.18.109:9100'}
        expected_alert_2 = {'instance': '135.248.18.109:9177',
                            'domain': 'instance-00000005'}
        self.assertEqual(expected_alert_1, observed_alert_1)
        self.assertEqual(expected_alert_2, observed_alert_2)

    def test_get_conf_resource(self):

        # Test setup
        driver = PrometheusDriver()
        alerts = self._generate_alerts()
        alert_1 = alerts[0]
        alert_2 = alerts[3]

        # Test Action
        observed_alert_1 = driver._get_conf_resource(alert_1)
        observed_alert_2 = driver._get_conf_resource(alert_2)

        # Test assertions
        expected_alert_1 = {'id': 'instance'}
        expected_alert_2 = {'instance_name': 'domain',
                            'host_id': 'instance'}
        self.assertEqual(expected_alert_1, observed_alert_1)
        self.assertEqual(expected_alert_2, observed_alert_2)

    def test_validate_ip(self):

        # Test setup
        driver = PrometheusDriver()
        ipv4_without_port = '1.1.1.1'
        ipv4_with_port = '1.1.1.1:1'
        invalid_ipv4 = '1.1'
        ipv6_without_port = '2001:db8::'
        ipv6_with_port = '[2001:db8::]:11'
        invalid_ipv6 = '2001:db8:'
        not_ip = 'not ip'

        # Test Action
        observed_ipv4_without_port = driver._validate_ip(ipv4_without_port)
        observed_ipv4_with_port = driver._validate_ip(ipv4_with_port)
        observed_ipv6_without_port = driver._validate_ip(ipv6_without_port)
        ipv6_with_port = driver._validate_ip(ipv6_with_port)

        # Test assertions
        self.assertIsNotNone(observed_ipv4_without_port)
        self.assertIsNotNone(observed_ipv4_with_port)
        self.assertIsNotNone(observed_ipv6_without_port)
        self.assertIsNotNone(ipv6_with_port)
        self.assertRaises(ValueError, driver._validate_ip, invalid_ipv4)
        self.assertRaises(ValueError, driver._validate_ip, invalid_ipv6)
        self.assertRaises(ValueError, driver._validate_ip, not_ip)

    @mock.patch('vitrage.datasources.prometheus.driver.'
                'PrometheusDriver.nova_client')
    @mock.patch('socket.gethostbyaddr')
    def test_enrich_event(self, mock_socket, mock_nova_client):

        # Test setup
        mock_nova_client.servers.list.return_value = None
        mock_socket.return_value = ['devstack-rocky-release-4']
        driver = PrometheusDriver()
        event = self._generate_event()

        # Test Action
        created_events = driver.enrich_event(event, PROMETHEUS_EVENT_TYPE)

        # Test assertions
        self._assert_event_equal(created_events, PROMETHEUS_EVENT_TYPE)

    def _assert_event_equal(self,
                            created_events,
                            expected_event_type):
        self.assertIsNotNone(created_events, 'No events returned')
        self.assertThat(created_events, matchers.HasLength(6),
                        'Expected 6 events')
        for event in created_events:
            self.assertEqual(expected_event_type,
                             event[DSProps.EVENT_TYPE])
            self.assertIsNotNone(event[PDProps.ENTITY_UNIQUE_PROPS])

    def _generate_alerts(self):
        event = self._generate_event()
        details = event[EProps.DETAILS]
        return details[PProps.ALERTS]

    @staticmethod
    def _generate_event(update_vals=None):
        generators = mock_driver.simple_prometheus_alarm_generators(
            update_vals=update_vals)

        return mock_driver.generate_sequential_events_list(generators)[0]



# Copyright 2015 - Alcatel-Lucent
# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log

from vitrage.common.constants import EntityCategory as ECategory
from vitrage.common.constants import GraphAction
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.transformer_base import TransformerBase
from vitrage.entity_graph.mappings.datasource_info_mapper import \
    DatasourceInfoMapper
from vitrage.entity_graph.processor import base as processor
from vitrage.entity_graph.processor import processor_utils as PUtils
from vitrage.entity_graph.processor.transformer_manager import \
    TransformerManager
from vitrage.graph import Direction

LOG = log.getLogger(__name__)


class Processor(processor.ProcessorBase):

    def __init__(self, e_graph=None):
        super(Processor, self).__init__()
        self.transformer_manager = TransformerManager()
        self.info_mapper = DatasourceInfoMapper()
        self._initialize_events_actions()
        self.entity_graph = e_graph

    def process_event(self, event):
        """Decides which action to run on given event

        Transforms the event into a tuple (vertex, neighbors,action).
        After transforming, it runs the correct action according to the
        action received from the transformer.

        :param event: The event to be processed
        :type event: Dictionary
        """

        LOG.debug('processor event:\n%s', event)

        self._enrich_event(event)
        entity = self.transformer_manager.transform(event)

        if entity.action not in self.actions:
            LOG.warning('Deprecated or unknown entity %s ignored', entity)
            return

        self._calculate_vitrage_aggregated_values(entity.vertex, entity.action)
        self._set_datasource_name(entity, event)
        self.actions[entity.action](entity.vertex, entity.neighbors)

    def create_entity(self, new_vertex, neighbors):
        """Adds new vertex to the entity graph

        Adds the entity to the entity graph, and connects it's neighbors

        :param new_vertex: The new vertex to add to graph
        :type new_vertex: Vertex

        :param neighbors: The neighbors of the new vertex
        :type neighbors: List
        """

        LOG.debug('Add entity to entity graph:\n%s', new_vertex)
        self._add_resource_details_to_alarm(new_vertex, neighbors)
        self.entity_graph.add_vertex(new_vertex)
        self._connect_neighbors(neighbors, set(), GraphAction.CREATE_ENTITY)

    def update_entity(self, updated_vertex, neighbors):
        """Updates the vertex in the entity graph

        Updates the in entity in the entity graph. In addition it removes old
        neighbor connections, and connects the new neighbors.

        :param updated_vertex: The vertex to be updated in the graph
        :type updated_vertex: Vertex

        :param neighbors: The neighbors of the updated vertex
        :type neighbors: List
        """

        LOG.debug('Update entity in entity graph:\n%s', updated_vertex)

        graph_vertex = self.entity_graph.get_vertex(updated_vertex.vertex_id)
        if graph_vertex and not PUtils.is_newer_vertex(graph_vertex,
                                                       updated_vertex):
            LOG.warning("Update event arrived later than expected - "
                        "graph_vertex: %s --- updated_vertex: %s",
                        graph_vertex, updated_vertex)
            return
        self._add_resource_details_to_alarm(updated_vertex, neighbors)
        PUtils.update_entity_graph_vertex(self.entity_graph,
                                          graph_vertex,
                                          updated_vertex)
        self._update_neighbors(updated_vertex, neighbors)

    def delete_entity(self, deleted_vertex, neighbors):
        """Deletes the vertex from the entity graph

        Marks the corresponding vertex and its edges as deleted

        :param deleted_vertex: The vertex to be deleted from the graph
        :type deleted_vertex: Vertex

        :param neighbors: The neighbors of the deleted vertex
        :type neighbors: List
        """

        LOG.debug('Delete entity from entity graph:\n%s', deleted_vertex)

        graph_vertex = self.entity_graph.get_vertex(deleted_vertex.vertex_id)

        if not graph_vertex:
            LOG.warning('Delete - vertex not found %s', deleted_vertex)
            return
        elif PUtils.is_deleted(graph_vertex):
            LOG.warning('Delete - vertex already deleted - '
                        "graph_vertex: %s --- deleted_vertex: %s",
                        graph_vertex, deleted_vertex)
            return
        elif not PUtils.is_newer_vertex(graph_vertex, deleted_vertex):
            LOG.warning("Delete event arrived later than expected - "
                        "graph_vertex: %s --- deleted_vertex: %s",
                        graph_vertex, deleted_vertex)
            return

        neighbor_vertices = self.entity_graph.neighbors(
            deleted_vertex.vertex_id)
        neighbor_edges = self.entity_graph.get_edges(
            deleted_vertex.vertex_id)

        for edge in neighbor_edges:
            PUtils.mark_deleted(self.entity_graph, edge)

        for vertex in neighbor_vertices:
            PUtils.delete_placeholder_vertex(self.entity_graph, vertex)

        PUtils.mark_deleted(self.entity_graph, deleted_vertex)

    def update_relationship(self, entity_vertex, neighbors):
        LOG.debug('Update relationship in entity graph:\n%s', neighbors)

        if entity_vertex:
            self.entity_graph.update_vertex(entity_vertex)
        for neighbor in neighbors:
            self.entity_graph.update_edge(neighbor.edge)

    def delete_relationship(self, updated_vertex, neighbors):
        LOG.debug('Delete relationship from entity graph:\n%s', neighbors)

        if updated_vertex:
            self.entity_graph.update_vertex(updated_vertex)
        for neighbor in neighbors:
            graph_edge = self.entity_graph.get_edge(neighbor.edge.source_id,
                                                    neighbor.edge.target_id,
                                                    neighbor.edge.label)
            if graph_edge:
                PUtils.mark_deleted(self.entity_graph, graph_edge)

    def remove_deleted_entity(self, vertex, neighbors):
        """Removes the deleted vertex from the entity graph

        Removes vertex that it's vitrage_is_deleted value is True

        :param vertex: The vertex to be removed from the graph
        :type vertex: Vertex

        :param neighbors: The neighbors of the deleted vertex
        :type neighbors: List - It's just a mock in this method
        """

        LOG.debug('Remove deleted entity from entity graph:\n%s', vertex)

        graph_vertex = self.entity_graph.get_vertex(vertex.vertex_id)

        if graph_vertex and PUtils.is_deleted(graph_vertex) and \
                PUtils.is_newer_vertex(graph_vertex, vertex):
            self.entity_graph.remove_vertex(vertex)
        else:
            LOG.warning("Remove deleted entity arrived on invalid resource: "
                        "deleted_vertex - %s, graph_vertex - %s",
                        vertex, graph_vertex)

    def _update_neighbors(self, vertex, neighbors):
        """Updates vertices neighbor connections

        1. Removes old neighbor connections
        2. connects the new neighbors.
        """

        (valid_edges, obsolete_edges) = self._find_edges_status(vertex,
                                                                neighbors)
        self._delete_old_connections(vertex, obsolete_edges)
        self._connect_neighbors(neighbors,
                                valid_edges,
                                GraphAction.UPDATE_ENTITY)

    def _connect_neighbors(self, neighbors, valid_edges, action):
        """Updates the neighbor vertex and adds the connection edges """
        if not neighbors:
            LOG.debug('connect_neighbors - nothing to do')
            return

        LOG.debug("Connect neighbors. Neighbors: %s, valid_edges: %s",
                  neighbors, valid_edges)
        for (vertex, edge) in neighbors:
            graph_vertex = self.entity_graph.get_vertex(vertex.vertex_id)
            if not graph_vertex or not PUtils.is_deleted(graph_vertex):
                if graph_vertex and not PUtils.is_newer_vertex(graph_vertex,
                                                               vertex):
                    LOG.warning("Neighbor update event arrived later than "
                                "expected - graph_vertex: %s --- "
                                "updated_vertex: %s", graph_vertex, vertex)
                else:
                    LOG.debug("Updates vertex: %s", vertex)
                    self._calculate_vitrage_aggregated_values(vertex, action)
                    PUtils.update_entity_graph_vertex(self.entity_graph,
                                                      graph_vertex,
                                                      vertex)
                if edge not in valid_edges:
                    LOG.debug("Updates edge: %s", edge)
                    self.entity_graph.update_edge(edge)
            else:
                LOG.debug("neighbor vertex wasn't updated: %s", vertex)

    def _delete_old_connections(self, vertex, obsolete_edges):
        """Deletes the "vertex" old connections

        Finds the old connections that are connected to updated_vertex,
        and marks them as deleted
        """
        if not obsolete_edges:
            LOG.debug('obsolete_edges - nothing to do')
            return

        LOG.debug('Delete old connections. Vertex:\n%s', vertex)
        # remove old edges and placeholder vertices if exist
        for edge in obsolete_edges:
            LOG.debug("Delete obsolete edge:\n%s", edge)
            PUtils.mark_deleted(self.entity_graph, edge)
            graph_ver = self.entity_graph.get_vertex(
                edge.other_vertex(vertex.vertex_id))
            PUtils.delete_placeholder_vertex(self.entity_graph, graph_ver)

    def _find_edges_status(self, vertex, neighbors):
        """Finds "vertex" valid and old connections

        Checks all the edges that are connected to the vertex in the entity
        graph, and finds which of them are old connections (edges that are no
        longer connected to those entities), and which are valid connections.
        """

        valid_edges = set()
        obsolete_edges = set()

        graph_neighbor_types = \
            PUtils.find_neighbor_types(neighbors)

        neighbor_edges = set(e for v, e in neighbors)
        for curr_edge in self.entity_graph.get_edges(vertex.vertex_id,
                                                     direction=Direction.BOTH):
            # check if the edge in the graph has a connection to the
            # same type of resources in the new neighbors list
            neighbor_vertex = self.entity_graph.get_vertex(
                curr_edge.other_vertex(vertex.vertex_id))

            is_connection_type_exist = PUtils.get_vertex_types(
                neighbor_vertex) in graph_neighbor_types

            if not is_connection_type_exist:
                valid_edges.add(curr_edge)
                continue

            if curr_edge in neighbor_edges:
                valid_edges.add(curr_edge)
            else:
                obsolete_edges.add(curr_edge)

        return valid_edges, obsolete_edges

    def _initialize_events_actions(self):
        self.actions = {
            GraphAction.CREATE_ENTITY: self.create_entity,
            GraphAction.UPDATE_ENTITY: self.update_entity,
            GraphAction.DELETE_ENTITY: self.delete_entity,
            GraphAction.UPDATE_RELATIONSHIP: self.update_relationship,
            GraphAction.DELETE_RELATIONSHIP: self.delete_relationship,
            GraphAction.REMOVE_DELETED_ENTITY: self.remove_deleted_entity,
        }

    def _calculate_vitrage_aggregated_values(self, vertex, action):
        LOG.debug("calculate event state")

        try:
            if action in [GraphAction.UPDATE_ENTITY,
                          GraphAction.DELETE_ENTITY,
                          GraphAction.CREATE_ENTITY]:
                graph_vertex = self.entity_graph.get_vertex(
                    vertex.vertex_id)
            elif action in [GraphAction.REMOVE_DELETED_ENTITY,
                            GraphAction.UPDATE_RELATIONSHIP,
                            GraphAction.DELETE_RELATIONSHIP]:
                return None
            else:
                LOG.error('unrecognized action: %s for vertex: %s',
                          action, vertex)
                return None

            self.info_mapper.vitrage_aggregate_values(vertex, graph_vertex)
        except Exception:
            LOG.exception("Calculate aggregated state failed.")

    def _enrich_event(self, event):
        attr = self.transformer_manager.get_enrich_query(event)
        if attr is None:
            return
        result = self.entity_graph.get_vertices(attr)
        event[TransformerBase.QUERY_RESULT] = result

    def _add_resource_details_to_alarm(self, vertex, neighbors):

        if not neighbors:
            return

        resource = None
        alarms = []

        if vertex.get(VProps.VITRAGE_CATEGORY) == ECategory.ALARM:
            alarms = [vertex]
            for neighbor in neighbors:
                if neighbor.vertex.get(VProps.VITRAGE_CATEGORY) ==\
                        ECategory.RESOURCE:
                    resource = neighbor.vertex
        elif vertex.get(VProps.VITRAGE_CATEGORY) == ECategory.RESOURCE:
            resource = vertex
            for neighbor in neighbors:
                if neighbor.vertex.get(VProps.VITRAGE_CATEGORY) == \
                        ECategory.ALARM:
                    alarms.append(neighbor.vertex)

        for alarm in alarms:
            if not resource:
                continue

            project_id = resource.get(VProps.PROJECT_ID)
            if not project_id:
                n_vertex = self.entity_graph.get_vertex(resource.vertex_id)
                project_id = n_vertex.get(VProps.PROJECT_ID) \
                    if n_vertex else None
            self.add_resource_details(
                alarm,
                r_id=resource.vertex_id,
                r_type=resource.get(VProps.VITRAGE_TYPE),
                r_project_id=project_id)

    @staticmethod
    def add_resource_details(alarm, r_id, r_type, r_project_id):
        alarm[VProps.VITRAGE_RESOURCE_ID] = r_id
        alarm[VProps.VITRAGE_RESOURCE_TYPE] = r_type
        if r_project_id:
            alarm[VProps.VITRAGE_RESOURCE_PROJECT_ID] = r_project_id

    @staticmethod
    def _set_datasource_name(entity, event):
        if entity.vertex and \
                (entity.action == GraphAction.CREATE_ENTITY or
                 entity.action == GraphAction.UPDATE_ENTITY):
            datasource_name = event.get(VProps.VITRAGE_DATASOURCE_NAME)
            if datasource_name:
                entity.vertex.properties[VProps.VITRAGE_DATASOURCE_NAME] = \
                    datasource_name



# Copyright 2016 - Nokia
# Copyright 2015 eNovance <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os


from keystoneauth1 import exceptions as ka_exception
from keystoneauth1 import loading as ka_loading
from keystoneauth1 import session
# noinspection PyPackageRequirements
from keystoneclient.v3 import client as ks_client_v3
from oslo_config import cfg
from oslo_log import log

CONF = cfg.CONF
LOG = log.getLogger(__name__)

CFG_GROUP = "service_credentials"


def get_session():
    """Get a vitrage service credentials auth session."""
    auth_plugin = ka_loading.load_auth_from_conf_options(CONF, CFG_GROUP)
    return ka_loading.load_session_from_conf_options(
        CONF, CFG_GROUP, auth=auth_plugin
    )


def get_client():
    """Return a client for keystone v3 endpoint."""
    sess = get_session()
    return ks_client_v3.Client(session=sess)


def get_service_catalog(client):
    return client.session.auth.get_access(client.session).service_catalog


def get_auth_token(client):
    return client.session.auth.get_access(client.session).auth_token


def get_client_on_behalf_user(auth_plugin):
    """Return a client for keystone v3 endpoint."""
    sess = session.Session(auth=auth_plugin)
    return ks_client_v3.Client(session=sess)


def create_trust_id(trustor_user_id, trustor_project_id, roles,
                    auth_plugin):
    """Create a new trust using the vitrage service user."""
    admin_client = get_client()
    trustee_user_id = admin_client.session.get_user_id()

    client = get_client_on_behalf_user(auth_plugin)
    trust = client.trusts.create(trustor_user=trustor_user_id,
                                 trustee_user=trustee_user_id,
                                 project=trustor_project_id,
                                 impersonation=True,
                                 role_names=roles)
    return trust.id


def delete_trust_id(trust_id, auth_plugin):
    """Delete a trust previously setup for the vitrage user."""
    client = get_client_on_behalf_user(auth_plugin)
    try:
        client.trusts.delete(trust_id)
    except ka_exception.NotFound:
        pass


OPTS = [
    cfg.StrOpt('region_name',
               default=os.environ.get('OS_REGION_NAME'),
               deprecated_name="region-name",
               help='Region name to use for OpenStack service endpoints.'),
    cfg.StrOpt('interface',
               default=os.environ.get(
                   'OS_INTERFACE', os.environ.get('OS_ENDPOINT_TYPE',
                                                  'public')),
               deprecated_name="os-endpoint-type",
               choices=('public', 'internal', 'admin', 'auth', 'publicURL',
                        'internalURL', 'adminURL'),
               help='Type of endpoint in Identity service catalog to use for '
                    'communication with OpenStack services.'),
]


def register_keystoneauth_opts():
    ka_loading.register_auth_conf_options(CONF, CFG_GROUP)
    ka_loading.register_session_conf_options(
        CONF, CFG_GROUP,
        deprecated_opts={'cacert': [
            cfg.DeprecatedOpt('os-cacert', group=CFG_GROUP),
            cfg.DeprecatedOpt('os-cacert', group="DEFAULT")]
        })



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


class AodhProperties(object):
    ALARM_ID = 'alarm_id'
    DESCRIPTION = 'description'
    ENABLED = 'enabled'
    EVENT = 'event'
    EVENT_RULE = 'event_rule'
    EVENT_TYPE = 'event_type'
    EVENT_RESOURCE_ID = 'traits.resource_id'
    NAME = 'name'
    STATE = 'state'
    PROJECT_ID = 'project_id'
    QUERY = 'query'
    REPEAT_ACTIONS = 'repeat_actions'
    RESOURCE_ID = 'resource_id'
    SEVERITY = 'severity'
    STATE_TIMESTAMP = 'state_timestamp'
    THRESHOLD = 'threshold'
    THRESHOLD_RULE = 'threshold_rule'
    GNOCCHI_RESOURCES_THRESHOLD = 'gnocchi_resources_threshold'
    TIMESTAMP = 'timestamp'
    TYPE = 'type'
    VITRAGE_ID = 'vitrage_id'
    DETAIL = 'detail'
    RULE = 'rule'
    GNOCCHI_RESOURCES_THRESHOLD_RULE = 'gnocchi_resources_threshold_rule'
    COMPOSITE_RULE = 'composite_rule'
    GNOCCHI_AGGREGATION_BY_RESOURCES_THRESHOLD_RULE = \
        'gnocchi_aggregation_by_resources_threshold_rule'
    GNOCCHI_AGGREGATION_BY_METRICS_THRESHOLD_RULE = \
        'gnocchi_aggregation_by_metrics_threshold_rule'
    USER_ID = 'user_id'
    STATE_REASON = 'state_reason'
    METRICS = 'metrics'
    TIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'


class AodhState(object):
    OK = 'ok'
    ALARM = 'alarm'
    INSUFFICIENT_DATA = 'insufficient_data'


class AodhEventType(object):
    CREATION = 'alarm.creation'
    RULE_CHANGE = 'alarm.rule_change'
    STATE_TRANSITION = 'alarm.state_transition'
    DELETION = 'alarm.deletion'


class AodhExtendedAlarmType(object):
    EVENT = 'event'
    VITRAGE = 'vitrage'
    THRESHOLD = 'threshold'
    GNOCCHI_RESOURCES_THRESHOLD = 'gnocchi_resources_threshold'
    COMPOSITE = 'composite'
    GNOCCHI_AGGREGATION_BY_METRICS_THRESHOLD =  \
        'gnocchi_aggregation_by_metrics_threshold'
    GNOCCHI_AGGREGATION_BY_RESOURCES_THRESHOLD = \
        'gnocchi_aggregation_by_resources_threshold'



# -*- encoding: utf-8 -*-
# Copyright 2015 - Alcatel-Lucent
# Copyright © 2014-2015 eNovance
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import base64
from collections import defaultdict
import copy
import hashlib
import itertools
import pickle
import random
import threading
import time
import zlib

from oslo_config import cfg
from oslo_log import log

import cProfile

LOG = log.getLogger(__name__)


def recursive_keypairs(d, separator='.'):
    # taken from ceilometer and gnocchi
    for name, value in sorted(d.items()):
        if isinstance(value, dict):
            for subname, subvalue in recursive_keypairs(value, separator):
                yield ('%s%s%s' % (name, separator, subname), subvalue)
        else:
            yield name, value


def opt_exists(conf_parent, opt):
    try:
        return conf_parent[opt]
    except cfg.NoSuchOptError:
        return False


def do_cprofile(func):
    def profiled_func(*args, **kwargs):
        profile = cProfile.Profile()
        try:
            profile.enable()
            result = func(*args, **kwargs)
            profile.disable()
            return result
        finally:
            profile.print_stats('cumulative')
    return profiled_func


def get_portion(lst, num_of_portions, portion_index):
    """Split a list into n slices and return the i'th slice

    :rtype: list
    """
    # First shuffle the items to create an even distribution
    # Use the same random seed to always get the same shuffle
    if num_of_portions < 1 or portion_index < 0 or \
            portion_index >= num_of_portions:
        raise Exception('Cannot get_portion %s %s',
                        num_of_portions,
                        portion_index)

    list_copy = copy.copy(lst)
    random.Random(0.5).shuffle(list_copy)

    portions = defaultdict(list)
    portion_indexes = range(num_of_portions)
    g = itertools.cycle(portion_indexes)
    for curr_item in list_copy:
        curr_portion = next(g)
        portions[curr_portion].append(curr_item)
    return portions[portion_index]


def spawn(target, *args, **kwargs):
    t = threading.Thread(target=target, args=args, kwargs=kwargs)
    t.daemon = True
    t.start()
    return t


def md5(obj):
    if isinstance(obj, tuple):
        obj = str([str(o) for o in obj])

    if isinstance(obj, str):
        return hashlib.md5(obj.encode('utf-8')).hexdigest()
    raise Exception('Unknown object for md5 %s' % obj)


def fmt(docstr):
    """Format a docstring for use as documentation in sample config."""
    # Replace newlines with spaces, as docstrings contain literal newlines that
    # should not be rendered into the sample configuration file (instead, line
    # wrappings should be applied automatically).
    docstr = docstr.replace('\n', ' ')

    # Because it's common for docstrings to begin and end with a newline, there
    # is now whitespace at the beginning and end of the documentation as a side
    # effect of replacing newlines with spaces.
    docstr = docstr.strip()

    return docstr


def timed_method(log_results=False, warn_above_sec=-1):
    def _decorator(function):
        def wrapper(*args, **kwargs):
            t1 = time.time()
            result = function(*args, **kwargs)
            t2 = time.time()
            if 0 < warn_above_sec < t2 - t1:
                LOG.warning(
                    'Function %s runtime crossed limit %s seconds.',
                    function.__name__, t2 - t1)
            elif log_results:
                LOG.info('Function %s timed %s', function.__name__, t2 - t1)
            return result
        return wrapper
    return _decorator


def compress_obj(obj, level=9):
    str_data = pickle.dumps(obj)
    data = base64.b64encode(zlib.compress(str_data, level))
    return data


def decompress_obj(blob):
    decoded_blob = base64.standard_b64decode(blob)
    str_data = zlib.decompress(decoded_blob)
    obj = pickle.loads(str_data)
    del decoded_blob
    del str_data
    return obj


def to_int(i):
    return int(i) if i is not None else i



# Copyright 2019 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.evaluator.template_validation.content.v1.get_param_validator \
    import GetParamValidator
from vitrage.tests.unit.evaluator.template_validation.content.base import \
    ValidatorTest


class ParametersValidatorTest(ValidatorTest):
    """Tests for the parameters validator of version 1

       All tests should succeed, as long as there is no get_param reference in
       the template itself
    """

    def test_validate_no_parameters(self):
        result = GetParamValidator.validate(
            template={'alarm_name': "Don't add a comment"}, actual_params=None)
        self._assert_correct_result(result)

    def test_validate_empty_parameters(self):
        result = GetParamValidator.validate(
            template={'alarm_name': '+2 for everybody'}, actual_params={})
        self._assert_correct_result(result)

    def test_validate_with_parameter(self):
        template = {'alarm_name': 'get_param(param1)'}
        result = \
            GetParamValidator.validate(template=template, actual_params={})
        self._assert_fault_result(result, 160)



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

OPTS = [
    cfg.StrOpt('notifier',
               default='vitrage.notifier.plugins.mistral.'
                       'mistral_notifier.MistralNotifier',
               help='mistral notifier class path',
               required=True),
]



# Copyright 2018 - Nokia Corporation
# Copyright 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noinspection PyPackageRequirements
from unittest import mock
import uuid

from datetime import datetime
# noinspection PyPackageRequirements
from http import client as httplib
from vitrage.tests.functional.api.v1 import FunctionalTest

HEADERS = {
    'Authorization': 'Basic dml0cmFnZTpwYXNzd29yZA==',
    'User-Agent': 'Alertmanager/0.15.0',
    'Host': '127.0.0.1:8999',
    'Content-Type': 'application/json'
}

EVENT_TYPE = 'prometheus.alarm'

VALID_TOKEN = uuid.uuid4().hex

PROJECT_ID = 'best_project'


class Role(object):
    pass


ROLES = Role()

ROLES.role_names = ['admin']

EVENT_DETAILS = {
    "status": "firing",
    "groupLabels": {
        "alertname": "HighInodeUsage"
    },
    "groupKey": "{}:{alertname=\"HighInodeUsage\"}",
    "commonAnnotations": {
        "mount_point": "/%",
        "description": "\"Consider ssh\"ing into the instance \"\n",
        "title": "High number of inode usage",
        "value": "96.81%",
        "device": "/dev/vda1%",
        "runbook": "troubleshooting/filesystem_alerts_inodes.md"
    },
    "alerts": [
        {
            "status": "firing",
            "labels": {
                "severity": "critical",
                "fstype": "ext4",
                "instance": "localhost:9100",
                "job": "node",
                "alertname": "HighInodeUsage",
                "device": "/dev/vda1",
                "mountpoint": "/"
            },
            "endsAt": "0001-01-01T00:00:00Z",
            "generatorURL": "http://devstack-rocky-4:9090/graph?g0.htm1",
            "startsAt": "2018-05-03T12:25:38.231388525Z",
            "annotations": {
                "mount_point": "/%",
                "description": "\"Consider ssh\"ing into the instance\"\n",
                "title": "High number of inode usage",
                "value": "96.81%",
                "device": "/dev/vda1%",
                "runbook": "troubleshooting/filesystem_alerts_inodes.md"
            }
        }
    ],
    "version": "4",
    "receiver": "vitrage",
    "externalURL": "http://devstack-rocky-4:9093",
    "commonLabels": {
        "severity": "critical",
        "fstype": "ext4",
        "instance": "localhost:9100",
        "job": "node",
        "alertname": "HighInodeUsage",
        "device": "/dev/vda1",
        "mountpoint": "/"
    }
}

ERR_MSG_MISSING_AUTH = 'The request you have made requires authentication.'

ERR_MSG_MISSING_VERSIONED_IDENTITY_ENDPOINTS = 'Authorization exception: ' \
                                               'Could not find versioned ' \
                                               'identity endpoints when ' \
                                               'attempting to authenticate. ' \
                                               'Please check that your ' \
                                               'auth_url is correct.'

ERR_MISSING_AUTH = {'error': {
    'code': httplib.UNAUTHORIZED,
    'title': httplib.responses[httplib.UNAUTHORIZED],
    'message': ERR_MSG_MISSING_AUTH,
}}

ERR_MISSING_VERSIONED_IDENTITY_ENDPOINTS = {'error': {
    'code': httplib.UNAUTHORIZED,
    'title': httplib.responses[httplib.UNAUTHORIZED],
    'message': ERR_MSG_MISSING_VERSIONED_IDENTITY_ENDPOINTS,
}}


class BasicAuthTest(FunctionalTest):

    def __init__(self, *args, **kwds):
        super(BasicAuthTest, self).__init__(*args, **kwds)
        self.auth = 'keystone'

    keystoneauth__identity = 'keystoneauth1.identity'

    @mock.patch('keystoneauth1.session.Session.get_token',
                return_value=VALID_TOKEN)
    @mock.patch('%s.base.BaseIdentityPlugin.get_project_id' %
                keystoneauth__identity,
                return_value=PROJECT_ID)
    @mock.patch('%s.generic.base.BaseGenericPlugin.get_auth_ref' %
                keystoneauth__identity,
                return_value=ROLES)
    @mock.patch('pecan.request')
    def test_header_parsing(self, req_mock, *args):
        resp = self.post_json('/event',
                              params={
                                  'time': datetime.now().isoformat(),
                                  'type': EVENT_TYPE,
                                  'details': EVENT_DETAILS
                              },
                              headers=HEADERS)
        req = resp.request
        self.assertEqual('Confirmed', req.headers['X-Identity-Status'])
        self.assertEqual(ROLES.role_names[0], req.headers['X-Roles'])
        self.assertEqual(PROJECT_ID, req.headers['X-Project-Id'])
        self.assertEqual(VALID_TOKEN, req.headers['X-Auth-Token'])
        self.assertEqual(1, req_mock.client.call.call_count)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_basic_mode_auth_wrong_authorization(self, *args):
        wrong_headers = HEADERS.copy()
        wrong_headers['Authorization'] = 'Basic dml0cmFnZTpwdml0cmFnZT=='
        resp = self.post_json('/event',
                              params={
                                  'time': datetime.now().isoformat(),
                                  'type': EVENT_TYPE,
                                  'details': EVENT_DETAILS
                              },
                              headers=wrong_headers,
                              expect_errors=True)
        self.assertEqual(httplib.UNAUTHORIZED, resp.status_code)
        self.assertDictEqual(ERR_MISSING_VERSIONED_IDENTITY_ENDPOINTS,
                             resp.json)

    def test_in_basic_mode_auth_no_header(self):
        resp = self.post_json('/event', expect_errors=True)

        self.assertEqual(httplib.UNAUTHORIZED, resp.status_code)
        self.assertDictEqual(ERR_MISSING_AUTH, resp.json)

    @mock.patch('keystoneauth1.identity.generic.password.Password')
    @mock.patch('keystoneauth1.session.Session')
    @mock.patch('pecan.request')
    def test_in_basic_mode_auth_success(self, req_mock, *args):
        resp = self.post_json('/event',
                              params={
                                  'time': datetime.now().isoformat(),
                                  'type': EVENT_TYPE,
                                  'details': EVENT_DETAILS
                              },
                              headers=HEADERS)

        self.assertEqual(1, req_mock.client.call.call_count)
        self.assertEqual(httplib.OK, resp.status_code)



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


class PropertiesElement(object):
    def __init__(self, properties=None):
        if properties is None:
            self.properties = {}
        self.properties = properties

    def __getitem__(self, key):
        """Get a property with 'value = element[key]'"""
        return self.properties[key]

    def __setitem__(self, key, value):
        self.properties[key] = value

    def __delitem__(self, key):
        """Delete a property with 'del(element[key])"""
        if key in self.properties:
            del self.properties[key]

    def __iter__(self):
        return self.properties.values()

    def get(self, k, d=None):
        return self.properties.get(k, d)

    def items(self):
        return self.properties.items()

    def copy(self):
        return PropertiesElement(self.properties.copy())


class Vertex(PropertiesElement):
    """Class Vertex

    A vertex is defined as follows:
    * vertex_id is a unique identifier
    * properties is a dictionary

    """

    def __init__(self, vertex_id, properties=None):
        """Create a Vertex instance

            :type vertex_id: str
            :type properties: dict
            :rtype: Vertex
            """
        super(Vertex, self).__init__(properties)
        if not vertex_id:
            raise AttributeError('Attribute vertex_id is missing')
        self.vertex_id = vertex_id

    def __hash__(self):
        return hash(self.vertex_id)

    def __repr__(self):
        return '{vertex_id : %s, properties : %s}' % \
               (self.vertex_id, self.properties)

    def __eq__(self, other):
        """Compare two vertices

        Example
        -------
        if vertex1 == vertex2:
            do something

        :type other: Vertex
        :rtype: bool
        """
        return vars(self) == vars(other) and \
            self.properties == other.properties

    def copy(self):
        return Vertex(vertex_id=self.vertex_id,
                      properties=self.properties.copy())


class Edge(PropertiesElement):
    """Class Edge represents a directional edge between two vertices

    An edge is defined as follows:
    * source_id is the first vertex id
    * target_id is the second vertex id
    * properties is a dictionary

    +---------------+    edge     +---------------+
    | source vertex |-----------> | target vertex |
    +---------------+             +---------------+

    """

    def __init__(self, source_id, target_id, label, properties=None):
        """Create an Edge instance

        :param source_id: source vertex id
        :type source_id: str

        :param target_id: target vertex id`
        :type target_id: str

        :param label:
        :type label: str

        :type properties: dict
        :rtype: Edge
        """
        super(Edge, self).__init__(properties)
        if not source_id:
            raise AttributeError('Attribute source_id is missing')
        if not target_id:
            raise AttributeError('Attribute target_id is missing')
        if not label:
            raise AttributeError('Attribute label is missing')
        self.source_id = source_id
        self.target_id = target_id
        self.label = label

    def __hash__(self):
        return hash('%s%s%s' % (self.source_id, self.target_id, self.label))

    def __repr__(self):
        return '{source_id : %s, target_id : %s, ' \
               'label = %s, properties : %s}' % (self.source_id,
                                                 self.target_id,
                                                 self.label,
                                                 self.properties)

    def __eq__(self, other):
        """Compare two edges

        Example
        -------
        if edge1 == edge2:
            do something

        :type other: Edge
        :rtype: bool
        """
        return vars(self) == vars(other) and \
            self.properties == other.properties

    def other_vertex(self, v_id):
        """If v_id == target_id return source_id, else return target_id

        :param v_id: the vertex id
        :return: the other vertex id
        """
        return self.source_id if self.target_id == v_id else self.target_id

    def has_vertex(self, v_id):
        return self.source_id == v_id or self.target_id == v_id

    def copy(self):
        return Edge(source_id=self.source_id,
                    target_id=self.target_id,
                    label=self.label,
                    properties=self.properties.copy())



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import EdgeLabel
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import GraphAction
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.alarm_transformer_base import AlarmTransformerBase
from vitrage.datasources.ceilometer import CEILOMETER_DATASOURCE
from vitrage.datasources.ceilometer.properties \
    import CeilometerEventType as CeilEventType
from vitrage.datasources.ceilometer.properties \
    import CeilometerProperties as CeilProps
from vitrage.datasources.ceilometer.properties \
    import CeilometerState as CeilState
from vitrage.datasources import transformer_base as tbase
from vitrage.datasources.transformer_base import Neighbor
from vitrage.datasources.transformer_base import TransformerBase
from vitrage.evaluator.actions.evaluator_event_transformer \
    import VITRAGE_DATASOURCE
import vitrage.graph.utils as graph_utils
from vitrage.utils import datetime as datetime_utils


class CeilometerTransformer(AlarmTransformerBase):

    # Event types which need to refer them differently
    GRAPH_ACTION_MAPPING = {
        CeilEventType.DELETION: GraphAction.DELETE_ENTITY,
    }

    def _create_snapshot_entity_vertex(self, entity_event):
        if _is_vitrage_alarm(entity_event):
            return self._create_merge_alarm_vertex(entity_event)
        return self._create_vertex(entity_event)

    def _create_update_entity_vertex(self, entity_event):
        if _is_vitrage_alarm(entity_event):
            return self._create_merge_alarm_vertex(entity_event)
        return self._create_vertex(entity_event)

    def _create_vertex(self, entity_event):
        metadata = {
            VProps.NAME: entity_event[CeilProps.NAME],
            VProps.SEVERITY: entity_event[CeilProps.SEVERITY],
            CeilProps.DESCRIPTION: entity_event[CeilProps.DESCRIPTION],
            CeilProps.ENABLED: entity_event[CeilProps.ENABLED],
            VProps.PROJECT_ID: entity_event.get(CeilProps.PROJECT_ID, None),
            CeilProps.REPEAT_ACTIONS: entity_event[CeilProps.REPEAT_ACTIONS],
            VProps.RESOURCE_ID: entity_event[CeilProps.RESOURCE_ID],
            'alarm_type': entity_event[CeilProps.TYPE]
        }

        # TODO(annarez): convert EVENT_TYPE to tuple
        if entity_event[CeilProps.TYPE] == CeilProps.EVENT:
            metadata[CeilProps.EVENT_TYPE] = \
                entity_event[CeilProps.EVENT_TYPE]

        elif entity_event[CeilProps.TYPE] == CeilProps.THRESHOLD:
            metadata[CeilProps.STATE_TIMESTAMP] = \
                entity_event[CeilProps.STATE_TIMESTAMP]

        vitrage_sample_timestamp = entity_event[DSProps.SAMPLE_DATE]

        update_timestamp = self._format_update_timestamp(
            CeilometerTransformer._timestamp(entity_event),
            vitrage_sample_timestamp)

        return graph_utils.create_vertex(
            self._create_entity_key(entity_event),
            vitrage_category=EntityCategory.ALARM,
            vitrage_type=entity_event[DSProps.ENTITY_TYPE],
            vitrage_sample_timestamp=vitrage_sample_timestamp,
            entity_id=entity_event[CeilProps.ALARM_ID],
            entity_state=self._get_alarm_state(entity_event),
            update_timestamp=update_timestamp,
            metadata=metadata)

    def _create_snapshot_neighbors(self, entity_event):
        return self._create_aodh_neighbors(entity_event)

    def _create_update_neighbors(self, entity_event):
        return self._create_aodh_neighbors(entity_event)

    def _create_aodh_neighbors(self, entity_event):
        graph_neighbors = entity_event.get(self.QUERY_RESULT, [])
        result = []
        for vertex in graph_neighbors:
            edge = graph_utils.create_edge(
                source_id=TransformerBase.uuid_from_deprecated_vitrage_id(
                    self._create_entity_key(entity_event)),
                target_id=vertex.vertex_id,
                relationship_type=EdgeLabel.ON)
            result.append(Neighbor(vertex, edge))
        return result

    def _create_merge_alarm_vertex(self, entity_event):
        """Handle an alarm that already has a vitrage_id

        This is a deduced alarm created in aodh by vitrage, so it already
        exists in the graph.
        This function will update the exiting vertex (and not create a new one)
        """
        metadata = {
            CeilProps.DESCRIPTION: entity_event[CeilProps.DESCRIPTION],
            VProps.PROJECT_ID: entity_event[CeilProps.PROJECT_ID],
        }
        vitrage_sample_timestamp = entity_event[DSProps.SAMPLE_DATE]
        update_timestamp = self._format_update_timestamp(
            CeilometerTransformer._timestamp(entity_event),
            vitrage_sample_timestamp)

        return graph_utils.create_vertex(
            self._create_entity_key(entity_event),
            vitrage_category=EntityCategory.ALARM,
            vitrage_type=VITRAGE_DATASOURCE,
            vitrage_sample_timestamp=vitrage_sample_timestamp,
            entity_id=entity_event.get(CeilProps.ALARM_ID),
            update_timestamp=update_timestamp,
            metadata=metadata)

    def _ok_status(self, entity_event):
        return entity_event[CeilProps.STATE] != CeilState.ALARM

    def _create_entity_key(self, entity_event):
        if _is_vitrage_alarm(entity_event):
            return entity_event.get(CeilProps.VITRAGE_ID)

        entity_type = entity_event[DSProps.ENTITY_TYPE]
        alarm_id = entity_event[CeilProps.ALARM_ID]
        return tbase.build_key((EntityCategory.ALARM, entity_type, alarm_id))

    @staticmethod
    def _timestamp(entity_event):
        return datetime_utils.change_time_str_format(
            entity_event[CeilProps.TIMESTAMP],
            CeilProps.TIME_FORMAT,
            tbase.TIMESTAMP_FORMAT)

    @staticmethod
    def get_enrich_query(event):
        affected_resource_id = event.get(CeilProps.RESOURCE_ID, None)
        if not affected_resource_id:
            return None
        return {VProps.ID: affected_resource_id}

    def get_vitrage_type(self):
        return CEILOMETER_DATASOURCE


def _is_vitrage_alarm(entity_event):
    return entity_event.get(CeilProps.VITRAGE_ID) is not None



# Copyright 2015 - Alcatel-Lucent
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.datasources.nagios import NAGIOS_DATASOURCE
from vitrage.datasources.nagios.transformer import NagiosTransformer
from vitrage.datasources.nova.host import NOVA_HOST_DATASOURCE
from vitrage.datasources.nova.host.transformer import HostTransformer
from vitrage.datasources.nova.instance import NOVA_INSTANCE_DATASOURCE
from vitrage.datasources.nova.instance.transformer import InstanceTransformer
from vitrage.datasources.nova.zone import NOVA_ZONE_DATASOURCE
from vitrage.datasources.nova.zone.transformer import ZoneTransformer
from vitrage.entity_graph.processor.transformer_manager import\
    TransformerManager
from vitrage.opts import register_opts
from vitrage.tests import base


class TransformerManagerTest(base.BaseTest):

    def setUp(self):
        super(TransformerManagerTest, self).setUp()
        self.cfg_fixture.config(group='datasources',
                                types=[
                                    NAGIOS_DATASOURCE,
                                    NOVA_HOST_DATASOURCE,
                                    NOVA_INSTANCE_DATASOURCE,
                                    NOVA_ZONE_DATASOURCE])

        for datasource in self.conf.datasources.types:
            register_opts(datasource, self.conf.datasources.path)

        self.manager = TransformerManager()

    def test_transformer_registration_nagios(self):
        self.assertIsInstance(self.manager.get_transformer
                              (NAGIOS_DATASOURCE), NagiosTransformer)

    def test_transformer_registration_nova_host(self):
        self.assertIsInstance(self.manager.get_transformer
                              (NOVA_HOST_DATASOURCE), HostTransformer)

    def test_transformer_registration_nova_instance(self):
        self.assertIsInstance(self.manager.get_transformer
                              (NOVA_INSTANCE_DATASOURCE), InstanceTransformer)

    def test_transformer_registration_nova_zone(self):
        self.assertIsInstance(self.manager.get_transformer
                              (NOVA_ZONE_DATASOURCE), ZoneTransformer)



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import abc


class Connection(object, metaclass=abc.ABCMeta):
    """Base class for storage system connections."""

    def __init__(self, url):
        pass

    @property
    def active_actions(self):
        return None

    @property
    def events(self):
        return None

    @property
    def templates(self):
        return None

    @property
    def graph_snapshots(self):
        return None

    @property
    def webhooks(self):
        return None

    @property
    def alarms(self):
        return None

    @property
    def edges(self):
        return None

    @property
    def changes(self):
        return None

    @abc.abstractmethod
    def disconnect(self):
        raise NotImplementedError('disconnect is not implemented')

    @abc.abstractmethod
    def clear(self):
        raise NotImplementedError('clear is not implemented')


class ActiveActionsConnection(object, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def create(self, active_action):
        """Create a new action.

        :type active_action: vitrage.storage.sqlalchemy.models.ActiveAction
        """
        raise NotImplementedError('create active action is not implemented')

    @abc.abstractmethod
    def update(self, active_action):
        """Update an existing action.

        :type active_action: vitrage.storage.sqlalchemy.models.ActiveAction
        """
        raise NotImplementedError('update active action is not implemented')

    @abc.abstractmethod
    def query(self,
              action_type=None,
              extra_info=None,
              source_vertex_id=None,
              target_vertex_id=None,
              action_id=None,
              score=None,
              trigger=None,
              ):
        """Yields a lists of active actions that match filters.

        :rtype: list of vitrage.storage.sqlalchemy.models.ActiveAction
        """
        raise NotImplementedError('query active actions is not implemented')

    @abc.abstractmethod
    def delete(self,
               action_type=None,
               extra_info=None,
               source_vertex_id=None,
               target_vertex_id=None,
               action_id=None,
               score=None,
               trigger=None,
               ):
        """Delete all active actions that match the filters."""
        raise NotImplementedError('delete active actions is not implemented')


class WebhooksConnection(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def create(self, webhook):
        """Create a new webhook.

        :type webhook:
        vitrage.storage.sqlalchemy.models.Webhook
        """
        raise NotImplementedError('create webhook is not implemented')

    @abc.abstractmethod
    def query(self,
              id=None,
              project_id=None,
              is_admin_webhook=None,
              url=None,
              headers=None,
              regex_filter=None,
              ):
        """Yields a lists of webhooks that match filters.

        :rtype: list of vitrage.storage.sqlalchemy.models.Webhook
        """
        raise NotImplementedError('query webhook is not implemented')

    @abc.abstractmethod
    def delete(self, id=None):
        """Delete all webhooks that match the filters."""
        raise NotImplementedError('delete webhook is not implemented')


class TemplatesConnection(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def create(self, template):
        """Add a new template.

        :type template: vitrage.storage.sqlalchemy.models.Template
        """
        raise NotImplementedError('Create Template not implemented')

    @abc.abstractmethod
    def update(self, uuid, var, value):
        """update existing template.

        :type template: vitrage.storage.sqlalchemy.models.Template
        """
        raise NotImplementedError('Update Template not implemented')

    @abc.abstractmethod
    def query(self, name=None, file_content=None,
              uuid=None, status=None, status_details=None,
              template_type=None):
        """Yields a lists of templates that match filters.

        :type: list of vitrage.storage.sqlalchemy.models.Template
        """
        raise NotImplementedError('Query Templates not implemented')

    @abc.abstractmethod
    def delete(self, name=None, uuid=None):
        """Delete existing template

        :type: list of vitrage.storage.sqlalchemy.models.Template
        """
        raise NotImplementedError('Delete Templates not implemented')


class EventsConnection(object, metaclass=abc.ABCMeta):
    def create(self, event):
        """Create a new event.

        :type event: vitrage.storage.sqlalchemy.models.Event
        """
        raise NotImplementedError('create event is not implemented')

    def update(self, event):
        """Update an existing event.

        :type event: vitrage.storage.sqlalchemy.models.Event
        """
        raise NotImplementedError('update event is not implemented')

    def query(self,
              event_id=None,
              collector_timestamp=None,
              payload=None,
              gt_collector_timestamp=None,
              lt_collector_timestamp=None):
        """Yields a lists of events that match filters.

        :rtype: list of vitrage.storage.sqlalchemy.models.Event
        """
        raise NotImplementedError('query events is not implemented')

    def delete(self, event_id=None):
        """Delete all events older than event_id"""
        raise NotImplementedError('delete events is not implemented')


class GraphSnapshotsConnection(object, metaclass=abc.ABCMeta):
    def create(self, graph_snapshot):
        """Create a new graph snapshot.

        :type graph_snapshot: vitrage.storage.sqlalchemy.models.GraphSnapshot
        """
        raise NotImplementedError('create graph snapshot not implemented')

    def update(self, graph_snapshot):
        """Update a graph snapshot.

        :type graph_snapshot: vitrage.storage.sqlalchemy.models.GraphSnapshot
        """
        raise NotImplementedError('update graph snapshot not implemented')

    def query(self):
        """Yields latest graph snapshot

        :rtype: vitrage.storage.sqlalchemy.models.GraphSnapshot
        """
        raise NotImplementedError('query graph snapshot not implemented')

    def delete(self):
        """Delete all graph snapshots taken until timestamp."""
        raise NotImplementedError('delete graph snapshots not implemented')


class AlarmsConnection(object, metaclass=abc.ABCMeta):
    def create(self, alarm):
        raise NotImplementedError('create alarm not implemented')

    def update(self, vitrage_id, key, val):
        raise NotImplementedError('update alarms not implemented')

    def end_all_alarms(self, end_time):
        raise NotImplementedError('end all alarms not implemented')

    def delete(self, timestamp=None):
        raise NotImplementedError('delete alarms not implemented')


class EdgesConnection(object, metaclass=abc.ABCMeta):
    def create(self, edge):
        raise NotImplementedError('create edge not implemented')

    def update(self, source_id, target_id, timestamp):
        raise NotImplementedError('update edge not implemented')

    def end_all_edges(self, end_time):
        raise NotImplementedError('end all edges not implemented')

    def delete(self):
        raise NotImplementedError('delete edges not implemented')


class ChangesConnection(object, metaclass=abc.ABCMeta):
    def create(self, change):
        raise NotImplementedError('create change not implemented')

    def add_end_changes(self, changes_to_add, end_time):
        raise NotImplementedError('add end changes not implemented')

    def delete(self):
        raise NotImplementedError('delete changes not implemented')
